from __future__ import print_function
from __future__ import print_function
import uuid
import numpy as np
import scipy.optimize
from ._population import Population
from pychemia.utils.mathematics import unit_vector


class RealFunction(Population):
    def __init__(self, function, ndim, limits, local_minimization=False):
        """
        Creates a simple population of points in R^N with
        N=ndim the dimensionality of space and
        using an univaluated function 'function'

        :param function: Routine to evaluate a function
        :param ndim: (int) Dimensions of function space
        :param limits: (numpy.ndarray)
        :return:
        """
        Population.__init__(self, 'Euclidean', 'global', use_mongo=False)
        self.tag = 'global'
        self.name = 'Real Function'
        self.function = function
        self.ndim = ndim
        if len(limits) == 2:
            self.limits = np.zeros((ndim, 2))
            self.limits[:, 0] = limits[0]
            self.limits[:, 1] = limits[1]
        else:
            self.limits = np.array(limits)
            assert (self.limits.shape == (2, ndim))

        self._members = []
        self._actives = []
        self.db = {}
        self.moves = {}
        self.pcdb = None
        self.local_minimization = local_minimization

    def __str__(self):
        ret = ' Euclidean Population\n\n'
        ret += ' Name:               %s\n' % self.name
        ret += ' Tag:                %s\n' % self.tag
        ret += '\n'
        ret += ' Members:            %d\n' % len(self.members)
        ret += ' Actives:            %d\n' % len(self.actives)
        ret += ' Evaluated:          %d\n' % len(self.evaluated)
        return ret

    def new_entry(self, data, active=True):
        ident = self.new_identifier()
        x = np.atleast_1d(data)
        self.db[ident] = {'x': x, 'fx': None}
        self.evaluate_entry(ident)
        if active:
            self.actives.append(ident)
        self.members.append(ident)
        return ident

    def add_random(self):
        x = np.random.random_sample(self.ndim)
        x = x * (self.limits[:, 1] - self.limits[:, 0]) + self.limits[:, 0]
        return self.new_entry(x), None

    def check_duplicates(self, ids):
        selection = self.ids_sorted(ids)
        ret = {}
        if len(ids) == 0:
            return ret
        for i in range(len(ids)):
            ident1 = selection[i]
            for j in range(i + 1, len(ids)):
                ident2 = selection[j]
                distance = self.distance(ident1, ident2)
                if distance < 1E-2:
                    ret[ident2] = ident1
        return ret

    def coordinate(self, i):
        return self.db[i]['x']

    def cross(self, ids):
        assert len(ids) == 2
        parent1 = self.coordinate(ids[0])
        parent2 = self.coordinate(ids[1])
        if self.ndim == 1:
            son1 = 0.5 * (parent1 + parent2)
            son2 = np.abs(parent1 - parent2)
        elif self.ndim == 2:
            son1 = np.array([parent1[0], parent2[1]])
            son2 = np.array([parent2[0], parent1[1]])
        else:
            split = np.random.randint(1, self.ndim - 1)
            son1 = np.concatenate((parent1[:split], parent2[split:]))
            son2 = np.concatenate((parent2[:split], parent1[split:]))
        new_ident1 = self.new_identifier()
        self.members.append(new_ident1)
        new_ident2 = self.new_identifier()
        self.members.append(new_ident2)
        self.db[new_ident1] = {'x': son1, 'fx': None}
        self.evaluate_entry(new_ident1)
        self.db[new_ident2] = {'x': son2, 'fx': None}
        self.evaluate_entry(new_ident2)
        if self.db[new_ident1]['fx'] > self.db[new_ident2]['fx']:
            return new_ident2, new_ident1
        else:
            return new_ident1, new_ident2

    def distance(self, imember, jmember):
        # The trivial metric
        x1 = self.db[imember]['x']
        x2 = self.db[jmember]['x']
        return np.linalg.norm(x2 - x1)

    def enable(self, ident):
        if ident not in self.actives:
            self.actives.append(ident)

    def evaluate(self):
        for ident in self.members:
            self.evaluate_entry(ident)

    def evaluate_entry(self, ident):
        x = self.db[ident]['x']
        if self.local_minimization:
            localmin = scipy.optimize.minimize(self.function, x)
            if self.is_inside(localmin.x):
                self.db[ident]['x'] = localmin.x
                self.db[ident]['fx'] = localmin.fun
            else:
                self.db[ident]['fx'] = self.function(x)
        else:
            self.db[ident]['fx'] = self.function(x)

    # def evaluator_daemon(self):
    #
    #     def worker(db, function, d):
    #         while True:
    #             for entry_id in db:
    #                 if db[entry_id]['fx'] is None:
    #                     self.evaluate_entry(entry_id)
    #             for entry_id in self.db:
    #                 d[entry_id] = self.db[entry_id]
    #             time.sleep(5)
    #
    #     manager = Manager()
    #     d = manager.dict()
    #
    #     p = Process(target=worker, args=(self.db, self.function, d))
    #     p.start()
    #     return p, d

    def is_evaluated(self, i):
        if i in self.db and self.db[i]['fx'] is not None:
            return True
        else:
            return False

    def from_dict(self, population_dict):
        pass

    def disable(self, ident):
        if ident in self.actives:
            self.actives.remove(ident)

    def get_values(self, selection):
        ret = {}
        for i in selection:
            ret[i] = self.value(i)
        return ret

    def member_str(self, imember):
        ret = '('
        for i in range(self.ndim):
            ret += '%5.2f' % self.db[imember]['x'][i]
            if i < self.ndim - 1:
                ret += ', '
            else:
                ret += ') -> '
        if self.value(imember) is not None:
            ret += '%5.2f' % self.value(imember)
        else:
            ret += 'None'
        return ret

    def move(self, imember, jmember, factor=0.2, in_place=False):
        """
        Moves imember in the direction of jmember
        If in_place is True the movement occurs on the
        same address as imember

        :param factor:
        :param imember:
        :param jmember:
        :param in_place:
        :return:
        """

        x1 = self.db[imember]['x']
        x2 = self.db[jmember]['x']
        newx = x1 + factor * (x2 - x1)
        if not in_place:
            new_ident = self.new_identifier()
            self.actives.append(new_ident)
            self.members.append(new_ident)
        else:
            new_ident = imember

        # print 'Moving',imember,'located at',x1
        if new_ident not in self.moves:
            # print 'No previous'
            self.moves[new_ident] = np.vstack((x1, newx))
        else:
            # print 'With previous'
            self.moves[new_ident] = np.vstack((self.moves[new_ident], newx))
        # print self.moves[new_ident]
        self.db[new_ident] = {'x': newx, 'fx': None}
        self.evaluate_entry(new_ident)
        return new_ident

    def is_inside(self, x):
        outside = False
        for i in range(self.ndim):
            if self.limits[i, 0] > x[i] or x[i] > self.limits[i, 1]:
                outside = True
                # print('New position out of limits', x, self.limits)
        return not outside

    def move_random(self, imember, factor=0.2, in_place=False, kind='move'):

        x = np.array(self.db[imember]['x'])
        newx = x
        outside = True
        while outside:
            dx = 2 * np.random.random_sample(self.ndim) - 1
            # print 'Random movement', dx, factor
            dx = unit_vector(dx)
            newx = x + factor * dx
            outside = not self.is_inside(newx)

        if not in_place:
            new_ident = self.new_identifier()
            self.actives.append(new_ident)
            self.members.append(new_ident)
        else:
            new_ident = imember
        self.db[new_ident] = {'x': newx, 'fx': None}
        self.evaluate_entry(new_ident)
        return new_ident

    @staticmethod
    def new_identifier():
        return str(uuid.uuid4())[-12:]

    def random_population(self, n):
        for i in range(n):
            self.add_random()

    def replace_failed(self):
        pass

    def recover(self):
        pass

    def save(self):
        wf = open('population.dat', 'w')
        for i in sorted(self.members):
            wf.write("%15s %12.3f %12.3f\n" % (i, self.db[i]['x'][0], self.db[i]['x'][1]))
        wf.close()
        wf = open('members.dat', 'w')
        for i in sorted(self.members):
            wf.write("%15s\n" % i)
        wf.close()

    def save_info(self):
        pass

    def set_value(self, i, y):
        self.db[i]['fx'] = y

    def str_entry(self, entry_id):
        ret = 'x = ['
        for i in self.db[entry_id]['x']:
            ret += '%7.2e ,' % i
        ret = ret[:-1] + '] f(x) = %7.2e' % self.db[entry_id]['fx']
        return ret

    def value(self, imember):
        return self.db[imember]['fx']

    def write_change(self, change):
        # print 'Write Change', change
        if change['change'] == 'promoted':
            self.db[change['from']]['from'] = change['from']
        else:
            self.db[change['to']]['from'] = change['from']

    @property
    def actives(self):
        return self._actives

    @property
    def members(self):
        return self._members

    def get_duplicates(self, ids):
        pass



import pychemia
import numpy as np


def test_composition():
    """
    Tests (pychemia.core.composition)                            :
    """
    pychemia.pcm_log.debug('Tests (pychemia.core.composition)')
    comp = pychemia.Composition('YBa2Cu3O7')
    assert sorted(comp.species) == [u'Ba', u'Cu', u'O', u'Y']
    assert (sorted(comp.symbols) == [u'Ba', u'Ba', u'Cu', u'Cu', u'Cu', u'O', u'O', u'O', u'O', u'O', u'O', u'O', u'Y'])
    assert (comp.formula == 'Ba2Cu3O7Y')
    assert (abs(comp.covalent_volume() - 285.185) < 1E-5)
    assert (comp.natom == 13)

    assert (comp.sorted_formula(sortby='hill') == 'Ba2Cu3O7Y')

    comp = pychemia.Composition('Na2Cl2')
    assert (comp.symbols == [u'Cl', u'Cl', u'Na', u'Na'])
    assert (pychemia.utils.periodic.valence(comp.symbols) == [7, 7, 1, 1])


def test_structure():
    """
    Tests (pychemia.core.structure)                              :
    """
    pychemia.pcm_log.debug('Tests (pychemia.core.structure)')
    a = 4.05
    b = a / 2
    fcc = pychemia.Structure(symbols=['Au'], cell=[[0, b, b], [b, 0, b], [b, b, 0]], periodicity=True)
    assert (fcc.natom == 1)
    fcc_copy = fcc.copy()
    fcc_copy.canonical_form()
    assert (abs(fcc.volume - fcc_copy.volume) < 1E-13)
    assert (np.linalg.norm(fcc_copy.lattice.angles - fcc.lattice.angles) < 1E-10)
    spc = fcc.supercell((3, 3, 3))
    assert (spc.natom == 27)



from __future__ import print_function
from .searcher import Searcher
from pychemia import pcm_log


class BeeAlgorithm(Searcher):
    def __init__(self, population, params=None, generation_size=32, stabilization_limit=10):
        """
        Implementation fo the Firefly algorithm for global minimization

        :param population:
        :param stabilization_limit:
        :return:
        """
        Searcher.__init__(self, population, generation_size, stabilization_limit)
        # Parameters
        self.ne = None  # Number of elite scout bees
        self.nre = None  # Number of elite foragers
        self.nb = None  # Number of best scout bees (excluding elites)
        self.nrb = None  # Number of best foragers
        self.ns = None  # Number of scouts (include elites,  best and others)
        self.n = None  # Number of bees (size of colony n=ne*nre + nb*nrb + ns )
        self.set_params(params)
        self.scouts_elite = None
        self.scouts_best = None
        self.scouts_others = None
        self.foragers = {}
        self.delta_change = 1

    def set_params(self, params):
        """
        Set all the parameters required by the bee algorithm,
        Check right values for all the parameters

        :param params:
        """
        self.n = self.generation_size  # Number of bees (size of colony n=ne*nre + nb*nrb + ns )
        self.ne = 3  # Number of elite scout bees
        self.nre = 3  # Number of elite foragers
        self.nb = 2  # Number of best scout bees (excluding elites)
        self.nrb = 2  # Number of best foragers
        # Number of scouts (include elites,  best and others)
        self.ns = self.n - self.ne * self.nre - self.nb * self.nrb
        if params is None:
            params = {}
        if 'nb' in params:
            self.nb = params['nb']
        if 'nrb' in params:
            self.nrb = params['nrb']
        if 'ne' in params:
            self.ne = params['ne']
        if 'nre' in params:
            self.nre = params['nre']
        assert (self.ns < self.n)

    def get_params(self):
        return {'nb': self.nb, 'ne': self.ne, 'nrb': self.nrb, 'nre': self.nre}

    def run_one(self):
        # Get a static selection of the values in the generation that are relaxed
        selection = self.population.ids_sorted(self.actives_in_generation)
        pcm_log.info('Size of selection : %d' % len(selection))

        if self.scouts_elite is None:
            print('First run', len(selection), self.ns)
            # During the first iteration all the selection are scouts
            # Lets choose from there the elite, the best and the actual
            # scouts for the next iterations
            if len(selection) >= self.ns:
                self.scouts_elite = list(selection[:self.ne])
                self.scouts_best = list(selection[self.ne:self.ne + self.nb])
                self.scouts_others = list(selection[self.ne + self.nb:self.ns - self.ne + self.nb])
                # Disable extra scouts (from an initial population)
                for entry_id in selection[self.ns:]:
                    self.population.disable(entry_id)
                self.print_status()

                # Add nre foragers around each elite scout
                for entry_id in self.scouts_elite:
                    self.pass_to_new_generation(entry_id, reason='Elite')
                    self.create_foragers(entry_id, self.nre)
                self.print_status()

                # Add nrb foragers around each best scout
                for entry_id in self.scouts_best:
                    self.pass_to_new_generation(entry_id, reason='Best')
                    self.create_foragers(entry_id, self.nrb)
                self.print_status()

                for u in range(self.generation_size - len(self.get_generation(self.current_generation + 1))):
                    ident, origin = self.population.add_random()
                    self.population.disable(ident)
                    self.generation[ident] = [self.current_generation + 1]
                    self.scouts_others.append(ident)
                    print('Added to other scouts', ident)
                self.print_status()

            else:
                pass
                # Number of scouts insufficient, increase their number with more
                # random members
                # for i in range(selection, self.ns):
                #    entry_id=self.population.add_random()
                #    print 'Population raised:',entry_id
        else:
            # New iterations look into each patch and see witch bees are on each patch
            print('Foragers %d' % len(self.foragers), self.foragers)
            print('Elite    %d' % len(self.scouts_elite), self.scouts_elite)
            print('Best     %d' % len(self.scouts_best), self.scouts_best)
            print('Others   %d' % len(self.scouts_others), self.scouts_others)

            self.process_scouts(self.scouts_elite, selection)
            self.process_scouts(self.scouts_best, selection)

            dead_bees = 0
            for j in list(self.scouts_others):
                if j not in selection:
                    # Consider a dead bee
                    self.scouts_others.remove(j)
                    dead_bees += 1
            scouts_others_alive = self.population.ids_sorted(self.scouts_others)
            if len(scouts_others_alive) > 0:
                best_scout = scouts_others_alive[0]
                worst_best = self.population.ids_sorted(self.scouts_best)[-1]
                if self.population.value(best_scout) < self.population.value(worst_best):
                    self.scouts_best.remove(worst_best)
                    self.scouts_best.append(best_scout)
                    self.population.disable(worst_best)

            # Add nre foragers around each elite scout
            for entry_id in self.scouts_elite:
                self.pass_to_new_generation(entry_id, reason='Elite')
                self.foragers[entry_id] = []
                self.create_foragers(entry_id, self.nre)
            self.print_status()

            # Add nrb foragers around each best scout
            for entry_id in self.scouts_best:
                self.pass_to_new_generation(entry_id, reason='Best')
                self.foragers[entry_id] = []
                self.create_foragers(entry_id, self.nrb)
            self.print_status()

            for i in self.scouts_others:
                self.population.disable(i)
            self.print_status()

            # for i in range(self.ns - self.ne - self.nb):
            for u in range(self.generation_size - len(self.get_generation(self.current_generation + 1))):
                ident, origin = self.population.add_random()
                self.population.disable(ident)
                self.generation[ident] = [self.current_generation + 1]
                self.scouts_others.append(ident)
                print('Added to other scouts', ident)

            print('After run_one:')
            self.print_status()

    def create_foragers(self, scout, n):
        for i in range(n):
            forager = self.population.move_random(scout, factor=self.delta_change, in_place=False, kind='move')
            print('For scout:', scout, ' new forager: ', forager)
            self.population.disable(forager)
            self.generation[forager] = [self.current_generation + 1]
            if scout not in self.foragers:
                self.foragers[scout] = [forager]
            else:
                self.foragers[scout].append(forager)
                # change={'to': scout}
                # self.write_change(forager, change)

    def process_scouts(self, scouts, selection):
        # Removing dead bees from log
        for ibee in list(scouts):
            print(ibee, len(self.foragers[ibee]))
            dead_bees = 0
            for j in list(self.foragers[ibee]):
                if j not in selection:
                    # Consider a dead bee
                    self.foragers[ibee].remove(j)
                    dead_bees += 1

        for ibee in list(scouts):
            foragers_alive = self.population.ids_sorted(self.foragers[ibee])
            print('Foragers that survived:', foragers_alive)
            if len(foragers_alive) > 0:
                best_forager = foragers_alive[0]
                if self.population.value(best_forager) < self.population.value(ibee):
                    # We found a better elite
                    scouts.remove(ibee)
                    scouts.append(best_forager)
                    self.population.disable(ibee)
                    for i in foragers_alive[1:]:
                        self.population.disable(i)

                else:
                    for i in foragers_alive[:]:
                        self.population.disable(i)



#!/usr/bin/env python
from __future__ import print_function
import os
import sys
import pychemia
import argparse

if __name__ == "__main__":

    description = """Create a surface for a given structure along the given miller indices and for a given
    number of layers"""

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-m', '--miller', nargs=3,
                        default=(0, 0, 1), metavar='N', type=int,
                        help='Miller indices (default: 0 0 1)')
    parser.add_argument('-l', '--layers',
                        default=1, metavar='N', type=int,
                        help='Number of layers (default: 1)')
    parser.add_argument('-i', '--input',
                        default='POSCAR', metavar='filename', type=str,
                        help='Input file (default: POSCAR)')
    parser.add_argument('-o', '--output',
                        default='POSCARsurf', metavar='filename', type=str,
                        help='Output file (default: POSCARsurf)')

    args = parser.parse_args()

    inputfile = args.input
    outputfile = args.output
    miller = tuple(args.miller)
    layers = args.layers

    if not os.path.isfile(inputfile):
        print('File not found: ', input)
        sys.exit(1)

    print('INPUT:  ', inputfile)
    print('OUTPUT: ', outputfile)
    print('MILLER: ', miller)
    print('LAYERS: ', layers)

    h = int(miller[0])
    k = int(miller[1])
    l = int(miller[2])

    structure = pychemia.code.vasp.read_poscar(inputfile)
    surf = pychemia.analysis.rotate_along_indices(structure, h, k, l, layers)

    print(surf)

    pychemia.code.vasp.write_poscar(surf, outputfile)

    symmetry = pychemia.crystal.CrystalSymmetry(structure)
    print('SpaceGroup (Original): ', symmetry.number())

    symmetry = pychemia.crystal.CrystalSymmetry(surf)
    print('SpaceGroup (Final)   : ', symmetry.number())



from __future__ import print_function
import os
import socket
import time
import pychemia
from multiprocessing import Process
from pychemia import pcm_log, HAS_PYMONGO

if HAS_PYMONGO:
    from pychemia.db import get_database


class FireballCollector:

    def __init__(self, db_settings, dbname, source_dir='/fireball-archive/data-mining', nconcurrent=1,
                 sleeping_time=120):
        """
        FireballCollector is a class to manage the execution of a function 'worker' for entries on a list of PyChemiaDB
         databases.
        The execution of each worker occurs directly on the machine executing the code, and the class controls the
        number of concurrent execution using the variable 'nconcurrent'

        :param db_settings: Common database settings for all the databases that will be processed by this evaluator.
                            All databases should share common settings like server name, user and password, ssl and
                            replicaset settings.
        :param dbname: List of database names, the common settings are stored in db_settings
        :param nconcurrent: Number of concurrent executions allowed. Their number usually should be the number of cores
                            available on the machine. (default: 1)
        :param sleeping_time: Time in seconds before try to search for new candidates to evaluate (default: 120 seconds)
        """
        self.db_settings = db_settings
        self.dbname = dbname
        self.source_dir = source_dir
        self.nconcurrent = nconcurrent
        self.sleeping_time = sleeping_time

        rf = open(self.source_dir+os.sep+'outputs.txt')
        self.outputs = [x.strip() for x in rf.readlines()]

    def worker(self, pcdb, path):

        files = os.listdir(path)
        properties = {}
        geo = None

        if 'fireball.in' in files:
            score = 'input'
            invars = pychemia.code.fireball.read_fireball_in(path + os.sep + 'fireball.in')
            properties['input'] = invars
            if 'param.dat' in files:
                score += '+param'
                param = pychemia.code.fireball.read_param(path + os.sep + 'param.dat')
                properties['param_dat'] = param
            if 'eigen.dat' in files:
                score += '+eigen'
                eigen = pychemia.code.fireball.read_eigen(path + os.sep + 'eigen.dat')
                properties['eigen_dat'] = eigen
            if 'answer.bas' in files:
                score += '+answer'
                geo = pychemia.code.fireball.read_geometry_bas(path + os.sep + 'answer.bas')

            for i in files:
                if i in self.outputs and os.path.isfile(path + os.sep + i):
                    rf = open(path + os.sep + i)
                    if "Welcome to FIREBALL" in rf.readline():
                        outputs = pychemia.code.fireball.read_fireball_stdout(path + os.sep + i)
                        properties['output'] = outputs
                    rf.close()

        if geo is not None:
            pcdb.insert(geo, properties)

    def process_directory(self, path, ret):

        files = os.listdir(path)
        score = True
        if 'fireball.in' in files and 'answer.bas' in files:
            for i in files:
                if os.path.isfile(path + os.sep + i):
                    rf = open(path + os.sep + i)
                    if "Welcome to FIREBALL" in rf.readline():
                        score = True
                    rf.close()

        if score:
            ret.append(path)

        for i in files:
            if os.path.isdir(path + os.sep + i):
                self.process_directory(path + os.sep + i, ret)

    def get_list_candidates(self):
        """
        Scan all databases looking for candidates for evaluation

        :return: A list of pairs, each pair contains the name of the database
                    and candidate identifier.
        """

        ret = []
        self.process_directory(self.source_dir, ret)
        return ret

    def is_evaluated(self, pcdb, path):
        if pcdb.entries.find({'properties.path': path}).count() > 0:
            return True
        else:
            return False

    def run(self):
        """
        Continuously search for suitable candidates to evaluation among a list of databases.

        :return:
        """

        procs = []
        ids_running = []

        # Creating a list to store the 'nconcurrent' running jobs
        for i in range(self.nconcurrent):
            procs.append(None)
            ids_running.append(None)

        # Main loop looking permanently for candidates for evaluation
        while True:

            to_evaluate = self.get_list_candidates()

            index = 0
            currently_evaluating = 0
            for j in range(self.nconcurrent):
                if procs[j] is not None and procs[j].is_alive():
                    currently_evaluating += 1
            print('Candidates to evaluate: %d  Candidates in evaluation: %d' % (len(to_evaluate), currently_evaluating))

            while index < len(to_evaluate):

                db_settings = dict(self.db_settings)
                # The first component of each pair in to_evaluate is the name of the database
                db_settings['name'] = self.dbname
                pcdb = get_database(db_settings)

                slot = None
                while True:
                    for j in range(self.nconcurrent):
                        if procs[j] is None or not procs[j].is_alive():
                            slot = j
                    if slot is None:
                        time.sleep(self.sleeping_time)
                    else:
                        break

                # The function is_evaluated needs two arguments, the database object and entry identifier and
                # must return a boolean to decide if the candidate should be evaluated.
                if self.is_evaluated(pcdb, to_evaluate[index]):
                    ids_running[slot] = to_evaluate[index]

                    # This is the actual call to the worker, it must be a function with 4 arguments:
                    # The database settings, the entry identifier, the working directory and arguments for the worker
                    procs[j] = Process(target=self.worker, args=(db_settings, to_evaluate[index]))
                    procs[j].start()
                else:
                    pcm_log.debug('Not evaluable: %s' % to_evaluate[index])
                index += 1
            time.sleep(self.sleeping_time)



"""
Useful routines to get physical data from octopus TD runs
"""

import os
from math import acos
import scipy.io.netcdf
from numpy import linalg, zeros, loadtxt, apply_along_axis, newaxis, polyfit, poly1d, dot, array
import pychemia


def get_last_iteration(dirname):
    """
    Get the last "td.xxxxxxx" found in dirname

    :param dirname: (str) Directory to take last information
    :return:
    """
    return sorted([int(x[3:]) for x in os.listdir(dirname) if x[:3] == 'td.' and x[-1].isdigit()])[-1]


def hirshfeld(dirname, iteration=None, spin=None):
    """
    Get the Hirshfeld charge analysis for all the atoms
    comparing for a given iteration and spin.

    Args:
        dirname:
           The directory where the octopus data was produced

        iteration:
           Iteration to make the difference, if absent the last iteration
           will be taken
        spin:
           Integer with the spin for which the data will be taken, if absent
           the addition of both spins will be returned

    Returns:
        A numpy array for the number of atoms with the Hirsfeld charges

    """
    if iteration is None:
        iteration = get_last_iteration(dirname)

    inp = pychemia.code.octopus.InputVariables(dirname + '/inp')
    if inp.variables['SpinComponents'] == 'spin_polarized':
        if spin is None:
            it, hirshfeld1 = hirshfeld(dirname, iteration, spin=1)
            it, hirshfeld2 = hirshfeld(dirname, iteration, spin=2)
            return it, hirshfeld1 + hirshfeld2
        else:
            sp = '-sp' + str(spin)
    else:
        sp = ''

    filename = dirname + '/td.' + str(iteration).zfill(7) + '/density' + sp + '-hirschfeld.dat'
    if os.path.isfile(filename):
        hirshfeld_value = loadtxt(filename)
    else:
        filename = dirname + '/td.' + str(iteration).zfill(7) + '/density' + sp + '-hirshfeld.dat'
        hirshfeld_value = loadtxt(filename)

    return iteration, hirshfeld_value


def deflection(dirname, iteration=None):
    """
    Get the deflection angle respect to the initial velocity
    for all the atoms and for a given iteration and spin.

    Args:
        dirname:
           The directory where the octopus data was produced

        iteration:
           Iteration to make the difference, if absent the last iteration
           will be taken

    Returns:
        A numpy array for the number of atoms with the Hirsfeld charges

    """

    if iteration is None:
        iteration = get_last_iteration(dirname)

    natom, iterations, times, positions, velocities, forces = pychemia.code.octopus.coordinates(dirname)
    angle = zeros(natom)
    univel = zeros((natom, 3))

    if iteration == 0:
        print('Error: To get the deflection some evolution is necessary (iteration!=0)')

    else:

        # Vector of displacement from (0 -> t)
        dis = positions[iteration] - positions[0]
        # Magnitude of the displacement
        mag = apply_along_axis(linalg.norm, 1, dis)
        # Unitary vectors of displacement
        unidis = dis / mag[:, newaxis]

        # Unitary vector of velocity
        vel = array(velocities[0])
        for i in range(natom):
            # Magnitude of the velocities
            mag[i] = linalg.norm(vel[i])
            # Unitary vectors of velocity
            if mag[i] != 0:
                univel[i] = (1.0 / mag[i]) * vel[i]
                angle[i] = acos(dot(univel[i], unidis[i]))
            else:
                angle[i] = 0.0

    return angle


def magnitude_velocity(dirname, iteration=None):
    """
    Magnitude of velocity for each atom

    :param dirname: (str) Directory to analyse velocities
    :param iteration: (int) Number of iteration to use
    :return:
    """

    if iteration is None:
        iteration = get_last_iteration(dirname)

    natom, iterations, times, positions, velocities, forces = pychemia.code.octopus.coordinates(dirname)

    magvel = apply_along_axis(linalg.norm, 1, velocities[iteration])

    return magvel


def change_kinetic_energy(dirname, iteration=None):
    """
    Change of Kinetic energy respect to initial value

    :param dirname: (str) Directory to read
    :param iteration: (int) Iteration number to read
    :return:
    """
    if iteration is None:
        iteration = get_last_iteration(dirname)

    iterations, times, energy_dict = pychemia.code.octopus.energy(dirname)

    return energy_dict['Kinetic'][iteration] - energy_dict['Kinetic'][0]


def value_in_sphere(dirname, keys, iteration=None, radius=4.5, spin=None):
    """
    Compute the value of a certain scalar quantity inside a sphere of a
    given radius

    :param dirname: (str) Directory to read
    :param keys: (list) Variables to read the scalar quantity
    :param iteration: (int) Number of iteration to read
    :param radius: (float) Radius of the sphere to integrate
    :param spin: (int) Spin contribution to read
    :return:
    """
    if iteration is None:
        iteration = get_last_iteration(dirname)

    natom, iterations, times, positions, velocities, forces = pychemia.code.octopus.coordinates(dirname)

    # Is spin polarized?
    inp = pychemia.code.octopus.InputVariables(dirname + '/inp')
    if inp.variables['SpinComponents'] == 'spin_polarized':
        if spin is None:
            it, cis1 = value_in_sphere(dirname, keys, iteration, radius, spin=1)
            it, cis2 = value_in_sphere(dirname, keys, iteration, radius, spin=2)
            return it, cis1 + cis2
        else:
            sp = '-sp' + str(spin)
    else:
        sp = ''

    cis = zeros(natom)
    for i in range(natom):

        filename = dirname + '/td.' + str(iteration).zfill(7)
        filename += '/density' + sp + '-analysis' + str(i + 1).zfill(3) + '.nc'

        if os.path.isfile(filename):
            data = scipy.io.netcdf.netcdf_file(filename, mmap=False)
            if (not keys[0] in data.variables) or (not keys[1] in data.variables):
                print('ERROR with ', keys)
            x = data.variables[keys[0]][:]
            if len(x) == 0:
                print('ERROR File empty:', filename)
            index = (abs(x - radius)).argmin()
            # print 'dirname',dirname
            # print 'index=',index
            # print 'radius (min,max)',X[0],X[-1]
            # Polynomial fit
            extra = 100
            if index + extra > len(x):
                final = len(x) - 1
            else:
                final = index + extra
            x = data.variables[keys[0]][index - extra:final]
            y = data.variables[keys[1]][index - extra:final]
            try:
                z = polyfit(x, y, 3)
                p = poly1d(z)
            except ValueError:
                print('Error fitting value for:', filename)

            # Get fit
            cis[i] = p(radius)
            data.close()
        else:
            print('ERROR Missing file:', filename)

    return iteration, cis


def charge_in_sphere(dirname, iteration=None, radius=4.5, spin=None):
    """
    Get the value at a given index of the
    data stored in the NetCDF file

    :param dirname: (str) Directory to read
    :param iteration: (int) Number of iteration to read
    :param radius: (float) Radius of the sphere to integrate
    :param spin: (int) Spin contribution to read
    :return:

    """

    keys = ['radeff', 'densit']

    return value_in_sphere(dirname, keys, iteration, radius, spin)


def dipolx_in_sphere(dirname, iteration=None, radius=4.5, spin=None):
    """
    Get the value at a given index of the
    data stored in the NetCDF file

    :param dirname: (str) Directory to read
    :param iteration: (int) Number of iteration to read
    :param radius: (float) Radius of the sphere to integrate
    :param spin: (int) Spin contribution to read
    :return:

    """

    keys = ['radeff', 'dipolX']

    return value_in_sphere(dirname, keys, iteration, radius, spin)



import itertools
import json
import socket
from multiprocessing import Pool

import numpy as np
import pymongo
from bson.objectid import ObjectId

from pychemia import Structure, HAS_PYMONGO
from pychemia.utils.periodic import atomic_symbols


class PyChemiaDB:
    def __init__(self, name='pychemiadb', host='localhost', port=27017, user=None, passwd=None, ssl=False,
                 replicaset=None):
        """
        Creates a MongoDB client to 'host' with 'port' and connect it to the database 'name'.
        Authentication can be used with 'user' and 'password'

        :param name: (str) The name of the database
        :param host: (str) The host as name or IP
        :param port: (int) The number of port to connect with the server (Default is 27017)
        :param user: (str) The user with read or write permissions to the database
        :param passwd: (str,int) Password to authenticate the user into the server

        """
        self.db_settings = {'name': name,
                            'host': host,
                            'port': port,
                            'user': user,
                            'passwd': passwd,
                            'ssl': ssl,
                            'replicaset': replicaset}
        self.name = name
        uri = 'mongodb://'
        if user is not None:
            uri += user
            if passwd is not None:
                uri += ':' + str(passwd)
            uri += '@'
        uri += host + ':' + str(port)
        if user is not None:
            uri += '/' + name
        if pymongo.version_tuple[0] == 2:
            self._client = pymongo.MongoClient(uri, ssl=ssl, replicaSet=replicaset)
        elif pymongo.version_tuple[0] == 3:
            self._client = pymongo.MongoClient(uri, ssl=ssl,
                                               ssl_cert_reqs=pymongo.ssl_support.ssl.CERT_NONE,
                                               replicaSet=replicaset)
        else:
            raise ValueError('Wrong version of pymongo')
        self.db = self._client[name]
        self.entries = self.db.pychemia_entries
        self.set_minimal_schema()

    def __str__(self):
        ret = ' Database Name:       %s\n' % self.name
        ret += ' Host:                %s\n' % self.db_settings['host']
        ret += ' Port:                %s\n' % self.db_settings['port']
        ret += ' User:                %s\n' % self.db_settings['user']
        ret += ' SSL:                 %s\n' % self.db_settings['ssl']
        return ret

    def save_json(self, filename='db_settings.json'):
        wf = open(filename, 'w')
        json.dump(self.db_settings, wf, sort_keys=True, indent=4, separators=(',', ': '))
        wf.close()

    def insert(self, structure, properties=None, status=None):
        """
        Insert a pychemia structure instance and properties
        into the database
        :param structure: (pychemia.Structure) An instance of Pychemia's Structure
        :param properties: (dict) Dictionary of properties
        :param status: (dict) Dictionary of status
        :return:
        """
        if status is None:
            status = {}
        if properties is None:
            properties = {}
        entry = {'structure': structure.to_dict, 'properties': properties, 'status': status}
        entry_id = self.entries.insert(entry)
        return entry_id

    def clean(self):
        self._client.drop_database(self.name)
        self.db = self._client[self.name]

    def update(self, entry_id, structure=None, properties=None, status=None):
        """
        Update the fields 'structure', 'properties' or 'status' for a given identifier 'entry_id'

        :param entry_id: (ObjectID, str)
        :param structure: (pychemia.Structure) Structure to update
        :param properties: (dict) Dictionary of properties to update
        :param status: (dict) Status dictionary

        :return: The identifier for the entry that was updated

        :rtype : ObjectId

        """

        assert (self.entries.find_one({'_id': entry_id}) is not None)
        entry = self.entries.find_one({'_id': entry_id})
        if structure is not None:
            if isinstance(structure, Structure):
                entry['structure'] = structure.to_dict
            elif isinstance(structure, dict):
                entry['structure'] = structure
            else:
                print('ERROR: Could not process the structure')
                print(type(structure))
        if properties is not None:
            entry['properties'] = properties
        if status is not None:
            entry['status'] = status

        self.entries.update({'_id': entry_id}, entry)
        return entry_id

    def find_AnBm(self, specie_a=None, specie_b=None, n=1, m=1):
        """
        Search for structures with a composition expressed as AnBm
        where one and only one between A or B is fixed and the numbers
        amounts n and m are both fixed

        :param specie_a: (str) atom symbol for the first specie
        :param specie_b: (str) atom symbol for the second specie
        :param n: number of atoms for specie 'a'
        :param m: number of atoms for specie 'b'
        :return:
        :return: (list) List of ids for all the structures that fulfill
                 the conditions
        """
        if specie_a is None and specie_b is None:
            raise ValueError("Enter a specie for A or B")
        elif specie_a is not None and specie_b is not None:
            raise ValueError("Only enter A or B, not both")
        elif specie_a is not None:
            atom_fixed = specie_a
            number_fixed = n
            number_unfixed = m
            assert (specie_a in atomic_symbols)
        else:
            atom_fixed = specie_b
            number_fixed = m
            number_unfixed = n
            assert (specie_b in atomic_symbols)

        ret = []
        for entry in self.entries.find({'nspecies': 2, 'formula': {'$regex': atom_fixed}}):
            comp = Structure.from_dict(entry['structure']).get_composition()
            if atom_fixed in comp.composition and comp.composition[atom_fixed] % number_fixed == 0:
                species = comp.species
                species.remove(atom_fixed)
                other_specie = species[0]
                # See if the fraction n/m is correct for A and B
                if number_unfixed * float(comp.composition[atom_fixed]) == number_fixed * float(
                        comp.composition[other_specie]):
                    ret.append(entry['_id'])
        return ret

    def find_composition(self, composition):
        """
        Search for structures with a pseudo-composition expressed as dictionary
        where symbols that are not atomic symbols such as A or X can be used to
        represent arbitrary atoms

        :return: (list) List of ids for all the structures that fulfill
                 the conditions
        """
        ret = []
        for entry in self.entries.find({'nspecies': len(composition)}):
            comp = Structure.from_dict(entry['structure']).get_composition()
            valid = True
            if sum(comp.composition.values()) % sum(composition.values()) != 0:
                valid = False
            vect1 = np.float64(np.sort(comp.composition.values()))
            vect2 = np.float64(np.sort(composition.values()))
            v12 = vect1 / vect2
            if not np.all(v12 == v12[0]):
                valid = False
            for ispecie in composition:
                if ispecie in atomic_symbols and ispecie not in comp.species:
                    valid = False

            if valid:
                print(comp.composition)
                ret.append(entry['_id'])
        return ret

    def get_structure(self, entry_id):
        """
        Return the structure in the entry with id 'entry_id'

        :rtype : Structure
        """
        entry = self.entries.find_one({'_id': entry_id})
        return Structure.from_dict(entry['structure'])

    def get_dicts(self, entry_id):
        """
        Return a tuple with the fields in an entry
        structure, properties and status

        :rtype : tuple
        """
        entry = self.entries.find_one({'_id': entry_id})
        structure_dict = entry['structure']
        if 'properties' in entry:
            properties = entry['properties']
        else:
            properties = None
        if 'status' in entry:
            status = entry['status']
        else:
            status = None

        return structure_dict, properties, status

    def get_entry(self, entry_id):
        return self.entries.find_one({'_id': entry_id})

    def is_locked(self, entry_id):
        """
        Return if a given entry is locked by someone
        evaluating the structure contained

        :rtype : bool
        """
        entry = self.entries.find_one({'_id': entry_id})

        if 'status' in entry and entry['status'] is not None and 'lock' in entry['status']:
            return True
        else:
            return False

    def lock(self, entry_id, name=None):
        if name is None:
            name = socket.gethostname()
        self.entries.update({'_id': entry_id}, {'$set': {'status.lock': name}})

    def unlock(self, entry_id, name=None):
        if name is None:
            self.entries.update({'_id': entry_id}, {'$unset': {'status.lock': 1}})
        else:
            self.entries.update({'_id': entry_id, 'status.lock': name}, {'$unset': {'status.lock': 1}})

    def set_minimal_schema(self):
        for entry_id in self.entries.find({'status': None}, {'status': 1}):
            print(entry_id)
            self.entries.update({'_id': entry_id['_id']}, {'$set': {'status': {}}})
        for entry_id in self.entries.find({'properties': None}):
            print(entry_id)
            self.entries.update({'_id': entry_id['_id']}, {'$set': {'properties': {}}})

    def create_static(self, field):

        for entry in self.entries.find({}):
            entry[field + '_static'] = entry[field]
            self.db.pychemia_entries.update({'_id': entry['_id']}, entry)

    def map_to_all(self, function, nparal=6):

        pool = Pool(processes=nparal)
        cursor = self.entries.find({}, no_cursor_timeout=True)
        print(cursor.count())
        entries = [entry['_id'] for entry in cursor]
        ret = pool.map(function, itertools.izip(itertools.repeat(self.db_settings), entries))
        cursor.close()
        return ret

    def replace_failed(self):
        for entry in self.entries.find({'status.relaxation': 'failed'}):
            st = self.get_structure(entry['_id'])
            comp = st.composition
            new_structure = Structure.random_cell(comp)
            self.entries.update({'_id': entry['_id']}, {'$unset': {'status.relaxation': 1,
                                                                   'status.target_forces': 1,
                                                                   'properties.energy': 1,
                                                                   'properties.forces': 1,
                                                                   'properties.stress': 1}})
            self.update(entry['_id'], structure=new_structure)

    def get_tags(self):
        entries = [x['status'].keys() for x in self.entries.find({}, {'status': 1})]
        ret = []
        for i in entries:
            for j in i:
                if j not in [u'target_forces', u'relaxation', u'lock'] and j not in ret:
                    ret.append(j)
        return ret


def get_database(db_settings):
    if 'host' not in db_settings:
        db_settings['host'] = 'localhost'
    if 'port' not in db_settings:
        db_settings['port'] = 27017
    if 'ssl' not in db_settings:
        db_settings['ssl'] = False
    if 'replicaset' not in db_settings:
        db_settings['replicaset'] = None

    if 'user' not in db_settings:
        pcdb = PyChemiaDB(name=db_settings['name'], host=db_settings['host'], port=db_settings['port'],
                          ssl=db_settings['ssl'], replicaset=db_settings['replicaset'])
    else:
        pcdb = PyChemiaDB(name=db_settings['name'], host=db_settings['host'], port=db_settings['port'],
                          user=db_settings['user'], passwd=db_settings['passwd'], ssl=db_settings['ssl'],
                          replicaset=db_settings['replicaset'])
    return pcdb


def object_id(entry_id):
    if isinstance(entry_id, str):
        return ObjectId(entry_id)
    elif isinstance(entry_id, ObjectId):
        return entry_id


def create_user(name, admin_name, admin_passwd, user_name, user_passwd, host='localhost', port=27017, ssl=False,
                replicaset=None):
    """
    Creates a new user for the database 'name'

    :param name: (str) The name of the database
    :param admin_name: (str) The administrator name
    :param admin_passwd: (str) Administrator password
    :param user_name: (str) Username for the database
    :param user_passwd: (str) Password for the user
    :param host: (str) Name of the host for the MongoDB server (default: 'localhost')
    :param port: (int) Port to connect to the MongoDB server (default: 27017)
    :param ssl: (bool) If True enable ssl encryption for communications to host (default: False)
    :param replicaset: (str, None) Identifier of a Replica Set

    """
    mc = pymongo.MongoClient(host=host, port=port, ssl=ssl, ssl_cert_reqs=pymongo.ssl_support.ssl.CERT_NONE,
                             replicaset=replicaset)
    mc.admin.authenticate(admin_name, admin_passwd)
    mc[name].add_user(user_name, user_passwd)
    return PyChemiaDB(name=name, user=user_name, passwd=user_passwd, host=host, port=port, ssl=ssl,
                      replicaset=replicaset)


def create_database(name, admin_name, admin_passwd, user_name, user_passwd, host='localhost', port=27017, ssl=False,
                    replicaset=None):
    return create_user(name, admin_name, admin_passwd, user_name, user_passwd, host, port, ssl, replicaset)


def has_connection():
    if not HAS_PYMONGO:
        return False
    import pymongo
    try:
        maxSevSelDelay = 1
        client = pymongo.MongoClient("localhost", serverSelectionTimeoutMS=maxSevSelDelay)
        client.server_info()  # force connection on a request as the
        # connect=True parameter of MongoClient seems
        # to be useless here
        return True
    except pymongo.errors.ServerSelectionTimeoutError as err:
        # do whatever you need
        print(err)
        return False



import getopt
import logging
import os
import sys

import numpy as np

from pychemia.code.vasp.task.elastic import mechanical_properties, elastic_moduli

logging.basicConfig(level=logging.DEBUG)


def usage(name):
    print("""
NAME
    %s

DESCRIPTION
    Compute the mechanical properties from a given 'OUTCAR' file

OPTIONS

    --help, -h
        Return information on the options and use of this script

    --outcar, -o <string> (Default: 'OUTCAR')
        OUTCAR file from which the mechanical properties will be computed

""" % os.path.basename(name))


def main(argv):
    try:
        opts, args = getopt.getopt(argv[1:], "ho:", ["help", "outcar="])
    except getopt.GetoptError:
        usage(argv[0])
        sys.exit(2)

    # Default Values
    outcar = 'OUTCAR'

    for opt, arg in opts:

        if opt in ("-h", "--help"):
            usage(argv[0])
            sys.exit()
        elif opt in ("-o", "--outcar"):
            outcar = arg

    tem = np.array(elastic_moduli(outcar)['total_elastic_moduli'])
    np.set_printoptions(precision=3)
    print(' Total Elastic Moduli')
    print(' --------------------\n')
    for irow in tem:
        for jelem in irow:
            if jelem == 0.0 or jelem == -0.0:
                sys.stdout.write(' %8d    ' % 0)
            else:
                sys.stdout.write(' %12.3f' % jelem)
        sys.stdout.write('\n')

    print('\n')
    mech = mechanical_properties(tem)
    print(" %20s %5s %12s %12s %12s" % ('Property'.ljust(20), 'Units', 'Voigt', 'Reuss', 'Average'))
    print(' ' + 65 * "-")
    for i in sorted(mech):
        print(" %20s %5s %12.3f %12.3f %12.3f" % (i.ljust(20), mech[i]['units'], mech[i]['Voigt'], mech[i]['Reuss'],
                                                  0.5 * (mech[i]['Voigt'] + mech[i]['Reuss'])))
    print(' ' + 65 * "-")


if __name__ == "__main__":
    main(sys.argv)



"""
Modules to manipulate input and output for several atomistic simulation codes.
Currently there are implementations for *ABINIT*, *DFTB+*, *Fireball*, an internal
LennardJones 'calculator', *Octopus* and *VASP*.

"""
from .codes import Codes
from .relaxator import Relaxator
from . import vasp
from . import dftb
from .lennardjones import LennardJones
from . import fireball
from pychemia import HAS_SCIPY
from . import sprkkr
from . import phonopy
from . import new_fireball

if HAS_SCIPY:
    from . import abinit

# __all__ = filter(lambda s: not s.startswith('_'), dir())



#!/usr/bin/env python

import logging
import os
import sys
import json
import getopt
import numpy as np
import pychemia
import pandas
from pychemia.utils.computing import get_int, get_float
from pychemia.code.vasp.task import ConvergenceKPointGrid, ConvergenceCutOffEnergy, IdealStrength

logging.basicConfig(level=logging.DEBUG)


def cleaner():
    files = ['DOSCAR', 'EIGENVAL', 'IBZKPT', 'OSZICAR', 'PCDAT', 'PROCAR', 'WAVECAR', 'XDATCAR']
    [os.remove(ifile) for ifile in files if os.path.exists(ifile)]


def usage(name):
    print("""
NAME

    %s

DESCRIPTION
    Computes the ideal strength of a material, the cell is deformed along the axis especified by 'expansion', the
    structure is enlarge of compress from ini_factor to fin_factor with intervals of delta_factor. A convergence
    of Kpoints and Cutoff energy are applied to preserve an accuracy of 'energy_tol'. The VASP code is used

OPTIONS

    --help, -h
        Return information on the options and use of this script

    --structure, -s <string> (Default: 'POSCAR')
        Structure for which the strength will be computed

    --output, -o <string> (Default: 'IdealStrength.json')
        File in JSON format where the results are stored

    --ini_factor, -i <float> (Default: 0.01)
        Initial factor for which the cell will be enlarge or compress. A value of 0.01 means increase by 1 percent

    --fin_factor, -f <float> (Default: 0.1)
        Final factor for enlarging or compressing the cell

    --delta_factor, -d <float> (Default: 0.01)
        Spacing between values of the factor

    --nparal, -n <int> (Default: 2)
        Number of MPI parallel processes for VASP

    --binary, -b <string> (Default: 'vasp')
        Path to the VASP executable

    --energy_tol, -e <float> (Default: 1E-3)
        Energy tolerance for three succesive values of the Kpoint grid or CutOff energy

    --target_forces, -t <float> (Default: 1E-3)
        Target forces for internal relaxation

    --expansion, -x <int> (Default: 111)
        Determines along which directions the lattice is deformed
        111 means along the 3 directions
        101 means along only along the 'a' and 'c' lattice lengths
""" % os.path.basename(name))


def main(argv):
    try:
        opts, args = getopt.getopt(argv[1:], "hs:o:i:f:d:n:b:e:x:t:", ["help", "structure=", "output=", "ini_factor=",
                                                                       "fin_factor=", "delta_factor", "nparal=",
                                                                       "binary=", "energy_tol=", "expansion=",
                                                                       "target_forces"])
    except getopt.GetoptError:
        usage(argv[0])
        sys.exit(2)

    if len(opts) == 0:
        usage(argv[0])
        sys.exit(2)

    # Default Values
    structure_file = 'POSCAR'
    delta_factor = 0.01
    ini_factor = 0.01
    fin_factor = 0.1
    output_file = 'IdealStrength.json'
    nparal = 2
    energy_tol = 1E-3
    target_forces = 1E-3
    binary = 'vasp'
    expansion = 111

    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage(argv[0])
            sys.exit()
        elif opt in ("-s", "--structure"):
            structure_file = arg
        elif opt in ("-o", "--output"):
            output_file = arg
        elif opt in ("-i", "--ini_factor"):
            ini_factor = get_float(arg)
        elif opt in ("-f", "--fin_factor"):
            fin_factor = get_float(arg)
        elif opt in ("-d", "--delta_factor"):
            delta_factor = get_float(arg)
        elif opt in ("-n", "--nparal"):
            nparal = get_int(arg)
        elif opt in ("-b", "--binary"):
            binary = arg
        elif opt in ("-e", "--energy_tol"):
            energy_tol = get_float(arg)
        elif opt in ("-t", "--target_forces"):
            target_forces = get_float(arg)
        elif opt in ("-x", "--expansion"):
            expansion = get_int(arg)
        elif opt in ("-s", "--structure"):
            structure_file = get_int(arg)

    expansion = [int(expansion / 100), int(expansion / 10) % 10, expansion % 10]
    if len(expansion) == 0:
        print(" ERROR: ini_factor, fin_factor and delta_factor are not creating a finite range of values")
        sys.exit(2)

    structure = pychemia.structure_from_file(structure_file)
    if structure is None:
        print(" ERROR: Invalid structure, no structure could be obtained from '%s'" % structure_file)
        sys.exit(2)

    if structure_file == 'POSCAR':
        os.rename('POSCAR', 'POSCAR_original')

    print("\n PyChemia Ideal Strenght")
    print(" =======================\n")
    print(" Scaling factors     : ", str(np.arange(ini_factor, fin_factor + 0.9 * delta_factor, delta_factor)))
    print(" Executable          : ", binary)
    print(" Energy tolerance    : ", energy_tol)
    print(" Target forces       : ", target_forces)
    print(" Expansion directions: ", str(expansion))
    print(" MPI number of procs : ", nparal)
    print(" Structure           :\n")
    print(structure)

    cleaner()
    print('\nConvergence of Cut-off Energy')
    print('-----------------------------\n')
    ce = ConvergenceCutOffEnergy(structure, energy_tolerance=energy_tol, binary=binary)
    if os.path.isfile('convergence_encut.json'):
        print('A previous convergence study was found, loading...')
        ce.load()
    if not ce.is_converge:
        ce.run(nparal)
        ce.save()
        ce.plot()
    encut = ce.best_encut
    print('ENCUT: ', encut)

    cleaner()
    print('\nConvergence of K-Point Grid')
    print('---------------------------\n')
    ck = ConvergenceKPointGrid(structure, workdir='.', binary=binary, encut=encut,
                               energy_tolerance=energy_tol, recover=True)
    if os.path.isfile('convergence_kpoints.json'):
        print('A previous convergence study was found, loading...')
        ck.load()
    if not ck.is_converge:
        ck.run(nparal)
        ck.save()
        ck.plot()
    kp = ck.best_kpoints
    kp_density = kp.get_density_of_kpoints(structure.lattice)
    print('KPOINT GRID', kp.grid)

    strenght = IdealStrength(structure, ini_factor, fin_factor, delta_factor, kp, kp_density, expansion, encut,
                             nparal, binary, target_forces, output_file)

    strenght.run(nparal)
    strenght.save()

    cleaner()


def plot_strain(filename):
    import matplotlib.pyplot as plt
    a = json.load(open(filename))
    df = pandas.DataFrame(a)

    def trace(x):
        return np.trace(np.array(x['stress'])[0])

    df['stress'] = df['output'].map(trace)
    plt.plot(df['factor'], df['stress'])
    plt.savefig(filename + '.pdf')


if __name__ == "__main__":
    main(sys.argv)



from __future__ import print_function
import os
import re
import itertools
import random
import numpy as np
from ._population import Population
from pychemia import pcm_log
from pychemia.code.abinit import InputVariables
from pychemia.code.abinit import AbinitOutput
from pychemia.utils.mathematics import gram_smith_qr, gea_all_angles, gea_orthogonal_from_angles, unit_vector
from pychemia.utils.serializer import generic_serializer


class OrbitalDFTU(Population):
    def __init__(self, name, input_path='abinit.in', num_electrons_dftu=None, num_indep_matrices=None,
                 connections=None):

        """
        This population is created with the purpose of global optimization of correlated orbitals 'd'
        and 'f'.


        The population is basically a collection of ABINIT inputs, the candidates have the same structure and
        uses the same input variables with exception of 'dmatpawu', the purpose of the
        population is to use global-population searchers to find the correlation matrices
        'dmatpawu' that minimizes the energy.

        The variable 'dmatpawu' is a list of numbers that can be arranged into N matrices.
        The matrices are 5x5 for 'd' orbitals and 7x7 for 'f' orbitals.

        :param name: The name of the 'PyChemiaDB' database created to stored the different
                        set of variables and the resulting output from abinit.
                     When using databases protected with username and password, the database
                        should be created independently and the database object must be use
                        as the 'name' argument

        :param abinit_input: The abinit input file, all the variables will be preserve for all new candidates, except
                        for 'dmatpawu' the only variable that changes.

        :param num_electrons_dftu: Example [5, 1, 5, 1]

        """
        # Call the parent class initializer to link the PychemiaDB that will be used
        Population.__init__(self, name, 'global')
        # Checking for existence of 'abinit.in'
        if not os.path.isfile(input_path):
            raise ValueError("Abinit input not found")
        # Reading the input variables and getting the structure
        self.input_path = input_path
        self.input = InputVariables(input_path)
        if 'dmatpawu' not in self.input.variables:
            raise ValueError("Abinit input file does not contain 'dmatpawu' variable")
        self.structure = self.input.get_structure()

        print('Orbital population:')
        print('Species [znucl]: %s' % self.input['znucl'])

        self.natpawu = 0
        print('Orbitals corrected:')
        for i in range(self.input['ntypat']):
            if self.input['lpawu'][i] == -1:
                print("%3s : False" % self.input['znucl'][i])
            else:
                print("%3s : True (l=%d)" % (self.input['znucl'][i], self.input['lpawu'][i]))
                self.natpawu += sum([1 for x in self.input['typat'] if x == i + 1])
        print('Number of atoms where DFT+U is applied: %d' % self.natpawu)

        # Computing the orbital that will be corrected
        # 2 (d-orbitals) or 3 (f-orbitals)
        self.maxlpawu = max(self.input['lpawu'])
        if self.maxlpawu == 2:
            print("Correlation of 'd' orbitals")
        elif self.maxlpawu == 3:
            print("Correlation of 'f' orbitals")

        # nsppol is the number of independent spin polarisations. Can take the values 1 or 2
        if self.input.has_variable('nsppol'):
            self.nsppol = self.input.get_value('nsppol')
        else:
            # Default from ABINIT
            self.nsppol = 1

        # nspinor it the umber of spinorial components of the wavefunctions
        if self.input.has_variable('nspinor'):
            self.nspinor = self.input.get_value('nspinor')
        else:
            self.nspinor = 1

        # nspden is the number of spin-density components
        if self.input.has_variable('nspden'):
            self.nspden = self.input.get_value('nspden')
        else:
            self.nspden = self.nsppol

        if self.nsppol == 1 and self.nspinor == 1 and self.nspden == 1:
            # Non-magnetic system (nsppol=1, nspinor=1, nspden=1):
            # One (2lpawu+1)x(2lpawu+1) dmatpawu matrix is given for each atom on which +U is applied.
            # It contains the "spin-up" occupations.
            self.nmatrices = self.natpawu
        elif self.nsppol == 2 and self.nspinor == 1 and self.nspden == 2:
            # Ferromagnetic spin-polarized (collinear) system (nsppol=2, nspinor=1, nspden=2):
            # Two (2lpawu+1)x(2lpawu+1) dmatpawu matrices are given for each atom on which +U is applied.
            # They contain the "spin-up" and "spin-down" occupations.
            self.nmatrices = 2 * self.natpawu
        elif self.nsppol == 1 and self.nspinor == 1 and self.nspden == 2:
            # Anti-ferromagnetic spin-polarized(collinear) system(nsppol=1, nspinor=1, nspden=2):
            # One(2lpawu + 1)x(2lpawu + 1) dmatpawu matrix is given for each atom on which +U is applied.
            # It contains the "spin-up" occupations.
            self.nmatrices = self.natpawu
        elif self.nsppol == 1 and self.nspinor == 2 and self.nspden == 4:
            # Non-collinear magnetic system (nsppol=1, nspinor=2, nspden=4):
            # Two (2lpawu+1)x(2lpawu+1) dmatpawu matrices are given for each atom on which +U is applied.
            # They contains the "spin-up" and "spin-down" occupations (defined as n_up=(n+|m|)/2 and n_dn=(n-|m|)/2),
            #    where m is the integrated magnetization vector).
            self.nmatrices = 2 * self.natpawu
        elif self.nsppol == 1 and self.nspinor == 2 and self.nspden == 1:
            # Non-collinear magnetic system with zero magnetization (nsppol=1, nspinor=2, nspden=1):
            # Two (2lpawu+1)x(2lpawu+1) dmatpawu matrices are given for each atom on which +U is applied.
            # They contain the "spin-up" and "spin-down" occupations;
            self.nmatrices = 2 * self.natpawu

        print('Variables controling the total number of matrices')
        print('nsppol : %d' % self.nsppol)
        print('nspinor: %d' % self.nspinor)
        print('nspden : %d' % self.nspden)
        print('Total number of matrices expected on dmatpawu: %d' % self.nmatrices)

        if num_electrons_dftu is None:
            abiinput = InputVariables(input_path)
            dmatpawu = np.array(abiinput['dmatpawu']).reshape(-1, self.ndim, self.ndim)
            params = dmatpawu2params(dmatpawu, 5)
            self.num_electrons_dftu = np.apply_along_axis(sum, 1, params['occupations'])
        else:
            self.num_electrons_dftu = np.array(num_electrons_dftu)
        print('Number of electrons for each correlation matrix: %s' % self.num_electrons_dftu)

        if num_indep_matrices is not None:
            self.num_indep_matrices = num_indep_matrices
        else:
            self.num_indep_matrices = self.nmatrices
        print('Number of independent matrices: %d' % self.num_indep_matrices)

        if connections is not None:
            self.connections = list(connections)
            if len(self.connections) != self.nmatrices:
                raise ValueError('Number of connections between matrices is not consistent with the number of matrices '
                                 'defined on dmatpawu')
            print('Connections: %s' % self.connections)
        else:
            self.connections = list(range(self.nmatrices))

    def __str__(self):
        ret = ' Population LDA+U\n\n'
        ret += ' Name:               %s\n' % self.name
        ret += ' Tag:                %s\n' % self.tag
        ret += ' Formula:            %s\n' % self.structure.formula
        ret += ' natpawu:            %d\n' % self.natpawu
        ret += ' nmatrices:          %d\n' % self.nmatrices
        ret += ' maxlpawu:           %d\n' % self.maxlpawu
        ret += ' num_indep_matrices: %s\n' % self.num_indep_matrices
        ret += ' connections:        %s\n' % self.connections

        ret += ' Members:            %d\n' % len(self.members)
        ret += ' Actives:            %d\n' % len(self.actives)
        ret += ' Evaluated:          %d\n' % len(self.evaluated)
        return ret

    @property
    def ndim(self):
        """
        Dimension of the matrices defined on dmatpawu, for 'd' orbitals is 5 for 'f' orbitals is 7

        :return:
        """
        return 2 * self.maxlpawu + 1

    def add_random(self):
        """
        Creates a new set of variables to reconstruct the dmatpawu

        matrix_i (integers) is a matrix natpawu x ndim with entries are 0 or 1
        matrix_d (deltas) is a matrix natpawu x ndim with entries are [0, 0.5)
        P (matrices) is a set of matrices natpawu x ndim x ndim
        Those three matrices allow to reconstruct the variable 'dmatpawu' used by ABINIT

        :return:
        """
        matrices_defined = []

        matrix_i = self.num_indep_matrices * [None]
        matrix_d = self.num_indep_matrices * [None]
        euler = self.num_indep_matrices * [None]

        for i in range(self.num_indep_matrices):
            nelect = self.num_electrons_dftu[i]
            val = [x for x in list(itertools.product(range(2), repeat=self.ndim)) if sum(x) == nelect]
            ii = val[np.random.randint(len(val))]
            dd = np.zeros(self.ndim)
            matrix_i[i] = list(ii)
            matrix_d[i] = list(dd)
            matrices_defined.append(self.connections[i])
            p = gram_smith_qr(self.ndim)
            euler[i] = gea_all_angles(p)

        data = {'euler_angles': euler, 'occupations': matrix_i, 'deltas': matrix_d,
                'num_matrices': self.num_indep_matrices, 'ndim': self.ndim}

        return self.new_entry(data), None

    def cross(self, ids):
        """
        Crossing algorithm used notably by GA to mix the information from several candidates
        This crossing algorithm is mixing the angles of two correlation matrices preserving the
        ordering of the atoms where the angles are applied. The occupations and deltas are also
        mixed independently of the euler angles.

        :param ids:
        :return:
        """
        assert len(ids) == 2

        properties1 = self.get_correlation_params(ids[0])
        properties2 = self.get_correlation_params(ids[1])

        euler_angles1 = properties1['euler_angles']
        euler_angles2 = properties2['euler_angles']

        occupations1 = properties1['occupations']
        occupations2 = properties2['occupations']

        deltas1 = properties1['deltas']
        deltas2 = properties2['deltas']

        newdata1 = {'euler_angles': [], 'occupations': [], 'deltas': [], 'num_matrices': self.num_indep_matrices,
                    'ndim': self.ndim}
        newdata2 = {'euler_angles': [], 'occupations': [], 'deltas': [], 'num_matrices': self.num_indep_matrices,
                    'ndim': self.ndim}

        for i in range(self.num_indep_matrices):
            rnd = random.randint(0, 1)
            if rnd == 0:
                newdata1['euler_angles'].append(euler_angles1[i])
                newdata2['euler_angles'].append(euler_angles2[i])
            else:
                newdata1['euler_angles'].append(euler_angles2[i])
                newdata2['euler_angles'].append(euler_angles1[i])
            rnd = random.randint(0, 1)
            if rnd == 0:
                newdata1['occupations'].append(occupations1[i])
                newdata2['occupations'].append(occupations2[i])
                newdata1['deltas'].append(deltas1[i])
                newdata2['deltas'].append(deltas2[i])
            else:
                newdata1['occupations'].append(occupations2[i])
                newdata2['occupations'].append(occupations1[i])
                newdata1['deltas'].append(deltas2[i])
                newdata2['deltas'].append(deltas1[i])

        entry_id = self.new_entry(newdata1)
        entry_jd = self.new_entry(newdata2)
        return entry_id, entry_jd

    def get_correlation_params(self, entry_id):
        return self.get_entry(entry_id, {'properties': 1}, with_id=False)['properties']

    def evaluate_entry(self, entry_id):
        """
        Evaluation externalized, no implemented

        :param entry_id:
        :return:
        """
        pass

    def from_dict(self, population_dict):
        return self.__init__(name=self.name,
                             input_path=population_dict['input_path'],
                             num_electrons_dftu=population_dict['num_electrons_dftu'],
                             connections=population_dict['connections'],
                             num_indep_matrices=population_dict['num_indep_matrices'])

    @property
    def to_dict(self):
        ret = super(self).to_dict
        ret['input_path'] = self.input_path
        ret['num_electrons_dftu'] = list(self.num_electrons_dftu)
        ret['num_indep_matrices'] = self.num_indep_matrices
        ret['connections'] = list(self.connections)
        return ret

    def new_entry(self, properties, active=True):
        """
        Creates a new entry on the population database from given data.

        :param properties: dictionary with 3 keys 'D' for deltas, 'I' for the integers
                        and eigen for the rotation matrix applied to the orbitals
        :param active: if True, the entry is enabled on the DB to be evaluated.
        :return:

        """
        status = {self.tag: active}
        properties['energy'] = None
        entry = {'structure': self.structure.to_dict, 'properties': properties, 'status': status}
        entry_id = self.insert_entry(entry)
        pcm_log.debug('Added new entry: %s with tag=%s: %s' % (str(entry_id), self.tag, str(active)))
        return entry_id

    def is_evaluated(self, entry_id):

        """
        One candidate is considered evaluated if it contains any finite value of energy on the properties.energy field

        :param entry_id:
        :return:
        """
        entry = self.get_entry(entry_id, {'_id': 0, 'properties': 1})
        if entry['properties']['energy'] is not None:
            return True
        else:
            return False

    def check_duplicates(self, ids):
        """
        For a given list of identifiers 'ids' checks the values for the function 'distance' and return a dictionary
          where each key is the identifier of a unique candidate and the value is a list of identifiers considered
          equivalents to it.

        :param ids:  List of identifiers for wich the check will be performed
        :return:
        """
        ret = {}
        print('Checking %d candidates' % len(ids))
        for i in range(len(ids)):
            for j in range(i + 1, len(ids)):
                if self.distance(ids[i], ids[j]) < 1E-3:
                    if ids[i] in ret:
                        ret[ids[i]].append(ids[j])
                    else:
                        ret[ids[i]] = [ids[j]]
        return ret

    def distance(self, entry_id, entry_jd):
        """
        Measure of distance for two entries with identifiers 'entry_id' and 'entry_jd'

        :param entry_id: Identifier of first entry
        :param entry_jd: Identifier of second entry
        :return:
        """

        properties1 = self.get_correlation_params(entry_id)
        properties2 = self.get_correlation_params(entry_jd)

        euler_angles1 = properties1['euler_angles']
        euler_angles2 = properties2['euler_angles']

        uvect1 = unit_vector(np.array(euler_angles1).flatten())
        uvect2 = unit_vector(np.array(euler_angles2).flatten())

        dist_euler = 1 - np.dot(uvect1, uvect2)

        return dist_euler

    def move_random(self, entry_id, factor=0.2, in_place=False, kind='move'):
        """
        Move one candidate with identifier 'entry_id' randomly with a factor
        given by 'factor'

        :param entry_id: Identifier of entry
        :param factor: Factor use to scale the randomness of change
        :param in_place: If True the candidate is changed keeping the identifier unchanged
        :param kind: Use when several algorithms are used for movement. One implemented here
        :return:
        """
        properties = self.get_correlation_params(entry_id)

        newdata = dict(properties)
        for i in range(self.num_indep_matrices):
            for j in range(self.ndim):
                perturbation = properties['euler_angles'][i][j] + 2*np.random.rand()*factor - factor
                if np.pi/2.0 > perturbation > -np.pi/2.0:
                    newdata['euler_angles'][i][j] = perturbation

        print(newdata)

        if in_place:
            return self.pcdb.db.pychemia_entries.update({'_id': entry_id},  {'$set': {'properties': newdata}})
        else:
            return self.new_entry(newdata, active=False)

    def move(self, entry_id, entry_jd, factor=0.2, in_place=False):
        """
        Move one candidate with identifier 'entry_id' in the direction of another candidate 'entry_jd'

        :param entry_id: Identifier of first entry (Origin)
        :param entry_jd: Identifier of second entry (Target)
        :param factor: Scale factor for change, 0 scale is the 'Origin' candidate, 1 is the 'Target' candidate
                        Intermediate values will change candidates accordingly
        :param in_place: If True the candidate is changed keeping the identifier unchanged
        :return:
        """

        properties1 = self.get_correlation_params(entry_id)
        properties2 = self.get_correlation_params(entry_jd)

        euler_angles1 = properties1['euler_angles']
        euler_angles2 = properties2['euler_angles']

        euler_angles_new = np.zeros((self.num_indep_matrices, int(self.ndim*(self.ndim-1)/2)))

        for i in range(self.num_indep_matrices):
            for j in range(int(self.ndim*(self.ndim-1)/2)):
                angle1 = euler_angles1[i][j]
                angle2 = euler_angles2[i][j]
                if angle1 < angle2:
                    if angle2 - angle1 < angle1 - angle2 + 2*np.pi:
                        direction = 1  # Forward
                        angle = angle2 - angle1
                    else:
                        direction = -1  # Backward
                        angle = angle1 - angle2 + 2*np.pi
                else:
                    if angle1 - angle2 < angle2 - angle1 + 2*np.pi:
                        direction = -1  # Backward
                        angle = angle1 - angle2
                    else:
                        direction = 1
                        angle = angle2 - angle1 + 2*np.pi

                euler_angles_new[i, j] = angle1 + direction*factor*angle
                if euler_angles_new[i, j] > np.pi:
                    euler_angles_new[i, j] -= -2*np.pi
                if euler_angles_new[i, j] < -np.pi:
                    euler_angles_new[i, j] += -2*np.pi

        newdata = dict(properties1)
        newdata['euler_angles'] = generic_serializer(euler_angles_new)

        if in_place:
            return self.pcdb.db.pychemia_entries.update({'_id': entry_id},  {'$set': {'properties': newdata}})
        else:
            return self.new_entry(newdata, active=False)

    def recover(self):
        pass

    def value(self, entry_id):
        """
        Return the energy value associated to the candidate with identifier 'entry_id'

        :param entry_id:
        :return:
        """
        entry = self.get_entry(entry_id, {'properties.energy': 1})
        return entry['properties']['energy']

    def str_entry(self, entry_id):
        entry = self.get_entry(entry_id)
        print(entry['properties']['O'], entry['properties']['D'])

    def get_duplicates(self, ids):
        return None

    def prepare_folder(self, entry_id, workdir='.', source_dir='.'):
        """
        Prepare directories for abinit execution

        :param entry_id:    bson.ObjectID of the entry that will be used for preparing the folder
        :param source_dir: (str) is the directory where 'abinit.files' and 'batch.pbs' should be present
                           those directories will be symbolically linked inside the individual work directories
        :param workdir: (str) Base work directory for abinit executions. Inside this directory, a set of subdirectories
                              will be created using the mongo ID as name.

        """
        # Individual workdir
        iworkdir = workdir+os.sep+str(entry_id)

        if not os.path.isdir(iworkdir):
            os.mkdir(iworkdir)

        for ifile in ['abinit.files', 'batch.pbs']:
            if os.path.lexists(iworkdir + os.sep + ifile):
                os.remove(iworkdir + os.sep + ifile)
            if not os.path.isfile(source_dir + os.sep + ifile):
                print('WARNIG: The file %s should be present on %s, symbolic links will be created pointing '
                      'to that location' % (ifile, source_dir))
            os.symlink(os.path.abspath(source_dir + os.sep + ifile), iworkdir + os.sep + ifile)

        abiinput = InputVariables(self.input_path)
        params = self.get_correlation_params(entry_id)
        dmatpawu = params2dmatpawu(params)
        abiinput['dmatpawu'] = list(dmatpawu.flatten())
        abiinput.write(iworkdir + os.sep + 'abinit.in')

    def collect_data(self, entry_id, workdir):

        old_properties = self.get_correlation_params(entry_id)

        if os.path.isfile(workdir + '/abinit.out'):
            ao = AbinitOutput(workdir + '/abinit.out')

            dmatpawu = get_final_dmatpawu(workdir + '/abinit.out')
            params = dmatpawu2params(dmatpawu, self.ndim)
            properties = dict(params)
            properties['original_params'] = old_properties

            if 'etot' in ao.get_energetics():
                energy = ao.get_energetics()['etot'][-1]
                print('Uploading energy data for %s' % entry_id)
                properties['energy'] = energy
                self.pcdb.db.pychemia_entries.update({'_id': entry_id}, {'$set': {'properties': properties}})
                return True
            else:
                return False
        else:
            return False


def params2dmatpawu(params):
    """
    Build the variable dmatpawu from the components stored in params

    :param params: dictionary with keys 'I', 'D' and 'eigen'
    :return:
    """
    # print(list(params.keys()))

    ndim = params['ndim']

    if 'num_matrices' in params:
        num_matrices = params['num_matrices']
    else:
        num_matrices = len(params['deltas'])

    occupations = np.array(params['occupations']).reshape(num_matrices, -1)
    deltas = np.array(params['deltas']).reshape(num_matrices, -1)
    euler_angles = np.array(params['euler_angles']).reshape(num_matrices, -1)

    print(num_matrices)
    print(euler_angles.shape)
    print(deltas.shape)
    print(occupations.shape)

    ret = np.zeros((num_matrices, ndim, ndim))

    for i in range(num_matrices):
        print(i)
        eigval = np.diag(occupations[i]).astype(float)
        for j in range(ndim):
            if eigval[j, j] == 0:
                eigval[j, j] += deltas[i, j]
            else:
                eigval[j, j] -= deltas[i, j]
        print(eigval)
        rotation = gea_orthogonal_from_angles(euler_angles[i])
        print(rotation)
        correlation = np.dot(np.dot(rotation, eigval), rotation.T)
        ret[i] = correlation
    return ret


def dmatpawu2params(dmatpawu, ndim):
    """
    Takes the contents of the variable 'dmatpawu' and return their components as a set of 'occupations', 'deltas'
    and 'euler_angles'
    The Euler angles is a ordered list of angles that can rebuild a rotation matrix 'R'
    The rotation matrix 'R' is ensured to be an element of SO(ndim), ie det(R)=1.
    When the eigenvectors return a matrix with determinant -1 a mirror on the first dimension is applied.
    Such condition has no effect on the physical result of the correlation matrix

    :param dmatpawu: The contents of the variable 'dmatpawu'. A list of number representing N matrices ndim x ndim
    :param ndim: ndim is 5 for 'd' orbitals and 7 for 'f' orbitals
    :return:
    """
    dm = np.array(dmatpawu).reshape((-1, ndim, ndim))
    num_matrices = dm.shape[0]
    eigval = np.array([np.linalg.eigh(x)[0] for x in dm])
    occupations = np.array(np.round(eigval), dtype=int)
    deltas = np.abs(eigval - occupations)
    rotations = np.array([np.linalg.eigh(x)[1] for x in dm])

    mirror = np.eye(ndim)
    mirror[0, 0] = -1

    for i in range(len(rotations)):
        if np.linalg.det(rotations[i]) < 0:
            rotations[i] = np.dot(rotations[i], mirror)

    euler_angles = np.array([list(gea_all_angles(p)) for p in rotations])

    return {'occupations': occupations, 'deltas': deltas, 'euler_angles': euler_angles, 'num_matrices': num_matrices,
            'ndim': ndim}


def get_pattern(params, ndim):
    """

    :param params:
    :param ndim:
    :return:
    """

    eigvec = np.array(params['eigvec']).reshape((-1, ndim, ndim))
    natpawu = len(eigvec)
    connection = np.zeros((natpawu, natpawu, ndim, ndim))

    # bb = np.dot(eigvec[0], np.linalg.inv(eigvec[3]))
    # connection = np.array(np.round(np.diagonal(bb)), dtype=int)

    iii = np.array(params['I'], dtype=int).reshape((-1, ndim))

    pattern = np.zeros((natpawu, natpawu))
    for i in range(natpawu):
        for j in range(i, natpawu):

            bb = np.dot(eigvec[0], np.linalg.inv(eigvec[3]))
            connection[i, j] = bb
            connection[j, i] = bb

            if np.all(np.array(iii[i] == iii[j])):
                pattern[i, j] = 1
                pattern[j, i] = 1
            else:
                pattern[i, j] = 0
                pattern[j, i] = 0

    return connection, pattern


def get_final_correlation_matrices_from_output(filename):
    rf = open(filename)
    data = rf.read()
    mainblock = re.findall('LDA\+U DATA[\s\w\d\-.=,>:]*\n \n', data)
    assert len(mainblock) == 1

    pattern = "For Atom\s*(\d+), occupations for correlated orbitals. lpawu =\s*([\d]+)\s*Atom\s*[\d]+\s*. Occ. " \
              "for lpawu and for spin\s*\d+\s*=\s*([\d\.]+)\s*Atom\s*[\d]+\s*. Occ. for lpawu and for " \
              "spin\s*\d+\s*=\s*([\d\.]+)\s*=> On atom\s*\d+\s*,  local Mag. for lpawu is[\s\d\w\.\-]*== Occupation " \
              "matrix for correlated orbitals:\s*Occupation matrix for spin  1\s*([\d\.\-\s]*)Occupation matrix " \
              "for spin  2\s*([\d\.\-\s]*)"
    ans = re.findall(pattern, mainblock[0])
    # print(ans)

    ret = []
    for i in ans:
        atom_data = {'atom number': int(i[0]),
                     'orbital': int(i[1]),
                     'occ spin 1': float(i[2]),
                     'occ spin 2': float(i[3])}
        matrix = [float(x) for x in i[4].split()]
        atom_data['matrix spin 1'] = list(matrix)
        matrix = [float(x) for x in i[5].split()]
        atom_data['matrix spin 2'] = list(matrix)
        ret.append(atom_data)
    return ret


def get_final_dmatpawu(filename):
    ret = get_final_correlation_matrices_from_output(filename)
    dmatpawu = []
    for i in ret:
        dmatpawu += i['matrix spin 1']
    return dmatpawu



import os
import subprocess
import pychemia

path = 'pychemia/test/data'


def test_abinit2xyz():
    """
    Test command abinit2xyz                                      :
    """
    if pychemia.HAS_SCIPY and pychemia.HAS_SCIENTIFIC:
        pychemia.info()
        print('Running abinit2xyz', pychemia.HAS_SCIPY, pychemia.HAS_SCIENTIFIC)
        script = 'bin/abi_abinit2xyz.py'
        args = path + '/abinit_03/abinit.in'
        subprocess.call(['python', script, args])
        mol1 = pychemia.io.xyz.load(path + '/abinit_03/abinit_DS1.xyz')
        mol2 = pychemia.io.xyz.load(path + '/abinit_03/abinit_DS2.xyz')
        assert (mol1.natom == 16)
        assert (mol2.natom == 16)
        os.remove(path + '/abinit_03/abinit_DS1.xyz')
        os.remove(path + '/abinit_03/abinit_DS2.xyz')
        os.remove(path + '/abinit_03/abinit.files')


def test_xyz2abinit():
    """
    Test command xyz2abinit                                      :
    """
    if pychemia.HAS_SCIPY and pychemia.HAS_SCIENTIFIC:
        script = 'bin/abi_xyz2abinit.py'
        args = path + '/xyz/chlorophyll.xyz'
        subprocess.call(['python', script, args])
        inp = pychemia.code.abinit.InputVariables(path + '/xyz/chlorophyll.xyz.in')
        assert (inp.get_value('natom') == 140)
        os.remove(path + '/xyz/chlorophyll.xyz.in')
        return inp


def test_plot_bonds():
    """
    Test command plot_bonds                                      :
    """
    if pychemia.HAS_SCIPY and pychemia.HAS_SCIENTIFIC:
        script = 'bin/abi_plot_bonds.py'
        arg1 = path + '/abinit_05/abinit.files:11'
        arg2 = path + '/abinit_05/abinit.files:21'
        subprocess.call(['python', script, arg1, arg2])
        os.remove("bonds.txt")
        os.remove("bonds.pdf")


def test_plot_hist():
    """
    Test command plot_hist                                       :
    """
    if pychemia.HAS_SCIPY and pychemia.HAS_SCIENTIFIC:
        script = 'bin/abi_plot_hist.py'
        arg1 = path + '/abinit_05/abinit.files'
        arg2 = '11'
        subprocess.call(['python', script, arg1, arg2])
        os.remove(path + '/abinit_05/abinit-o_DS11.pdf')



#!/usr/bin/env python
"""
This example shows how to automatize the execution
of ABINIT starting with just the input file "t44.in"

The abinit.files is written with the right
locations for the pseudopotentials and the results
are post-process using the output in NetCDF format
"abinit-o_OUT.nc"
"""

import os
import sys
import shutil
import tempfile
import subprocess
import pychemia

if pychemia.HAS_SCIPY and pychemia.HAS_SCIENTIFIC:
    import pychemia.code.abinit as pa

path = 'pychemia/test/data'


def test_example1():
    """
    Example of a simple calc                                     :
    """
    if pychemia.HAS_SCIPY and pychemia.HAS_SCIENTIFIC:
        workdir = tempfile.mkdtemp()
        print(workdir)
        creation(workdir)
        execution(workdir)
        ecut, etotal = datamining(workdir)
        print('ecut=', ecut)
        print('etotal=', etotal)
        assert ecut == 15.0
        assert abs(etotal + 4.199348203363531) < 0.001
        shutil.rmtree(workdir)


def which(program):
    """
    Search for the presence of an executable
    Found in: http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
    """
    import os

    def is_exe(filep):
        return os.path.isfile(filep) and os.access(filep, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for ipath in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(ipath, program)
            if is_exe(exe_file):
                return exe_file
    return None


def creation(filep):
    """
    Create the input file from the original and
    set the proper values for abinit.files
    """
    var = pa.InputVariables(path + os.sep + 'abinit_04' + os.sep + 't44.in')
    abifile = pa.AbiFiles(filep)
    abifile.set_input(var)
    abifile.set_psps('LDA', 'FHI')
    var.write(abifile.get_input_filename())
    print(abifile)
    print(var)
    abifile.create()


def execution(filep):
    """
    Execute ABINIT in the given path
    """
    abifile = open(filep + '/abinit.files')
    logfile = open(filep + '/abinit.log', 'w')
    cwd = os.getcwd()
    os.chdir(filep)
    if which('abinit') is not None:
        subprocess.call(['abinit'], stdin=abifile, stdout=logfile)
    else:
        print('The executable "abinit" is not in the PATH')
        print('Using the results of a previous calc')
    os.chdir(cwd)
    abifile.close()
    logfile.close()


def datamining(filep):
    """
    Read some output variables from abinit-o_OUT.nc
    """
    print('Extracting ecut and etotal')
    if os.path.isfile(filep + os.sep + 'abinit-o_OUT.nc'):
        data = pa.netcdf2dict(filep + os.sep + 'abinit-o_OUT.nc')
    else:
        data = pa.netcdf2dict(path + os.sep + 'abinit_04' + os.sep + 'abinit-o_OUT.nc')
    print(data)
    print(data['ecut'], data['etotal'])
    return data['ecut'], data['etotal']


if __name__ == '__main__':
    workdir = tempfile.mkdtemp()
    if len(sys.argv) == 1:
        test_example1()
    elif len(sys.argv) == 2:
        if sys.argv[1] == 'create':
            creation(workdir)
        if sys.argv[1] == 'execute':
            execution(workdir)
        elif sys.argv[1] == 'result':
            datamining(workdir)



#!/usr/bin/env python

import os
import pychemia
from pychemia.test import has_local_mongo


def test_popu_noncoll():
    """
    Tests (pychemia.population.NonCollinearMagMoms)              :
    """
    if not pychemia.HAS_PYMONGO:
        print('PyChemiaDB was disabled')
        return
    else:
        if not has_local_mongo():
            return

    source = 'pychemia/test/data/vasp_02'
    assert os.path.isfile(source + os.sep + 'INCAR')
    assert os.path.isfile(source + os.sep + 'POSCAR')
    popu = pychemia.population.NonCollinearMagMoms('test', source)
    popu.pcdb.clean()
    popu.random_population(16)

    assert len(popu) == 16
    popu.pcdb.clean()


if __name__ == '__main__':
    test_popu_noncoll()



import json
import os
import time
from pychemia import pcm_log
from pychemia.crystal import KPoints
from ..dftb import DFTBplus, read_detailed_out


class StaticCalculation:
    def __init__(self, structure, workdir, slater_path, waiting=False, kpoints=None, output_file='results.json',
                 max_scc_iterations=50):

        self.structure = structure
        self.workdir = workdir
        self.slater_path = slater_path
        self.waiting = waiting
        self.MaxSCCIterations = max_scc_iterations
        if isinstance(slater_path, str):
            self.slater_path = [slater_path]
        self.results = []
        self.output_file = output_file
        if kpoints is None:
            self.kpoints = KPoints.optimized_grid(self.structure.lattice, kp_density=10000, force_odd=True)
        else:
            self.kpoints = kpoints

    def run(self):

        dftb = DFTBplus()
        dftb.initialize(workdir=self.workdir, structure=self.structure, kpoints=self.kpoints)
        dftb.set_slater_koster(search_paths=self.slater_path)
        dftb.kpoints = self.kpoints

        if os.path.isfile('charges.bin'):
            os.remove('charges.bin')

        for mixer in ['Broyden', 'Anderson', 'DIIS', 'Simple']:

            dftb.basic_input()
            dftb.hamiltonian['MaxSCCIterations'] = self.MaxSCCIterations
            dftb.hamiltonian['Mixer'] = {'name': mixer}
            if os.path.isfile('charges.bin'):
                dftb.hamiltonian['ReadInitialCharges'] = True

            ret = None
            dftb.set_static()
            dftb.set_inputs()
            dftb.run()
            if self.waiting:
                dftb.runner.wait()
            while True:
                if dftb.runner is not None and dftb.runner.poll() is not None:
                    pcm_log.info('Execution completed. Return code %d' % dftb.runner.returncode)
                    filename = dftb.workdir + os.sep + 'detailed.out'
                    ret = read_detailed_out(filename)
                    print('Mixer= %10s  Total_energy= %9.3f  iSCC= %4d  SCC_error= %9.3E' % (mixer,
                                                                                             ret['total_energy'],
                                                                                             ret['SCC']['iSCC'],
                                                                                             ret['SCC']['SCC_error']))
                    break
                time.sleep(10)

            if ret['SCC']['iSCC'] < self.MaxSCCIterations:
                break

            if ret is not None:
                self.results.append({'Mixer'
                                     'kp_grid': self.kpoints.grid,
                                     'iSCC': ret['SCC']['iSCC'],
                                     'Total_energy': ret['total_energy'],
                                     'SCC_error': ret['SCC']['SCC_error']})

    def save_json(self):

        wf = open(self.output_file, 'w')
        json.dump(self.results, wf)
        wf.close()



"""
Visualization of Anthracene
"""

import os
import time
import pychemia
import pychemia.external.pymatgen

path = 'pychemia/test/data'


def test_anthracene():
    """
    Visualization of anthracene         :
    """
    print(os.getcwd())
    assert (os.path.isdir(path))
    filename = path + '/xyz/anthracene.xyz'
    assert (os.path.isfile(filename))
    mol = pychemia.io.xyz.load(filename)
    assert (mol.natom == 24)
    assert (mol.is_crystal is False)
    assert (mol.is_periodic is False)
    fig = mol.plot()
    time.sleep(5)
    fig.remove()
    fig.stop()


def test_chlorophyll():
    """
    Visualization of chlorophyll        :
    """
    print(os.getcwd())
    assert (os.path.isdir(path))
    filename = path + '/xyz/chlorophyll.xyz'
    assert (os.path.isfile(filename))
    mol = pychemia.io.xyz.load(filename)
    assert (mol.natom == 140)
    assert (mol.is_periodic is False)
    assert (mol.is_crystal is False)
    fig = mol.plot()
    time.sleep(5)
    fig.remove()
    fig.stop()


def test_gold():
    """
    Visualization of Gold FCC lattice   :
    """
    a = 4.05
    b = a / 2
    fcc = pychemia.Structure(symbols=['Au'], cell=[[0, b, b], [b, 0, b], [b, b, 0]], periodicity=True)
    assert (fcc.natom == 1)
    assert (fcc.is_periodic is True)
    assert (fcc.is_crystal is True)
    fig = fcc.plot()
    time.sleep(5)
    fig.remove()
    fig.stop()



import re
import os
import numpy as np
from scipy.io import netcdf_file


class AbinitOutput:
    def __init__(self, filename='abinit.out'):

        self.filename = filename
        rf = open(self.filename)
        self.data = rf.read()

    def reload(self):
        rf = open(self.filename)
        self.data = rf.read()

    def get_energetics(self):

        ret = re.findall('ETOT\s*([\d]+)\s*([.E\d\-+]+)\s*([.E\d\-+]+)\s*([.E\d\-+]+)\s*([.E\d\-+]+)\n',
                         self.data)
        if not ret:
            return None
        else:
            ret = [[float(x) for x in y] for y in ret]
            ret = np.array(ret)
            energetics = {'iter': list(ret[:, 0]),
                          'etot': list(ret[:, 1]),
                          'deltaEh': list(ret[:, 2]),
                          'residm': list(ret[:, 3]),
                          'nres2': list(ret[:, 4])}
            return energetics

    @property
    def is_finished(self):
        return "Calculation completed." in self.data

    def get_occupation_matrix(self):

        occ_block = re.findall(r"=== For Atom[\s\w\d\-.,=>:]*\n \n \n", self.data)
        if not occ_block:
            return None
        occs = re.findall('Occupation matrix for spin[\s\d\w]*([\s\d\-.]*)', occ_block[0])
        atoms = [int(x) for x in re.findall('For Atom([\s\w\d]*)', occ_block[0])]
        spins = [int(x) for x in re.findall('Occupation matrix for spin([\s\w\d]*\n)', occ_block[0])]

        ret = []
        n = len(occs)
        if 2 * len(atoms) == len(spins) and len(spins) == n:
            for i in range(n):
                atom = atoms[(i + 1) / 2 + (i + 1) % 2 - 1]
                spin = spins[i]
                occ_matrix = [float(x) for x in occs[i].strip().split()]
                ret.append({'atom': atom, 'spin': spin, 'occ_matrix': occ_matrix})
        return ret

    def get_dmatpawu(self):
        om = self.get_occupation_matrix()
        min_atom = min([x['atom'] for x in om])
        max_atom = max([x['atom'] for x in om])
        ret = []
        for i in range(min_atom, max_atom + 1):
            for iom in om:
                if iom['atom'] == i and iom['spin'] == 1:
                    ret.append(iom['occ_matrix'])
        return np.array(ret)

    @staticmethod
    def read_output_netcdf(filename):
        return netcdf2dict(filename)


def netcdf2dict(filename):
    """
    Read a NetCDF file and create a python dictionary with
    numbers or lists for each variable

    Args:
        filename:
            NetCDF filename
    """
    if not os.path.isfile(filename):
        print('ERROR: No such file: ', filename)
        return None
    ret = {}
    netcdfile = netcdf_file(filename, 'r', mmap=False)
    for ii in netcdfile.variables.keys():
        ret[ii] = netcdfile.variables[ii][:]
    netcdfile.close()

    for i in ret:
        if ret[i].dtype == np.dtype('>f8'):
            ret[i] = [round(x, 11) for x in ret[i].flatten()]
        elif ret[i].dtype == np.dtype('>i4'):
            ret[i] = [int(x) for x in ret[i].flatten()]

    for i in ret:
        if len(ret[i]) == 1:
            ret[i] = ret[i][0]

    return ret



import itertools

import numpy as np

from pychemia import Composition, Structure


class SplitMatch:
    def __init__(self, structure1, structure2):

        assert structure1.nsites == structure2.nsites

        self.structures = [None, None]
        self.structures[0] = structure1
        self.structures[1] = structure2
        self.cut_planes = [None, None]
        self.split_sites = [None, None]
        self.ss_tags = [None, None]

    def get_minimal_splitting(self):

        delta_distance = 0.1
        while True:
            for i in range(2):
                self.cut_planes[i] = get_cut_planes(self.structures[i], delta_distance=delta_distance)
                self.split_sites[i] = get_simple_split_sites(self.structures[i], self.cut_planes[i])
                if len(self.split_sites[i]) == 0:
                    delta_distance *= 0.9
                    print('delta distance reduced to', delta_distance)
            if len(self.split_sites[0]) > 0 and len(self.split_sites[1]) > 0:
                # Select two compatible splits (that preserve the number of atoms)
                has_groups = False
                self.ss_tags[0] = None
                self.ss_tags[1] = None
                grp0 = None
                grp1 = None
                trial = 0
                while not has_groups:
                    rnd = np.random.randint(len(self.split_sites[0]))
                    grp0 = self.split_sites[0].keys()[rnd]
                    for grp1 in self.split_sites[1]:
                        trial += 1
                        print('Selected Groups:', grp0, grp1)
                        if len(self.split_sites[0][grp0]) == len(self.split_sites[1][grp1]):
                            has_groups = True
                            break
                    if trial > 10:
                        print('Bad splitting of sites, decreasing distance')
                        grp0 = None
                        grp1 = None
                        break

                if grp0 is not None and grp1 is not None:
                    self.ss_tags[0] = grp0
                    self.ss_tags[1] = grp1
                    break
                else:
                    delta_distance *= 0.9

    def get_simple_match(self):

        self.get_minimal_splitting()

        # grp1 and grp2 are tuples where the first value is the dimension of cut and the second one
        # is the second one is the index for the plane cut
        grp0 = self.ss_tags[0]
        grp1 = self.ss_tags[1]

        idim1 = grp0[0]
        idim2 = grp1[0]
        cp1 = grp0[1]
        cp2 = grp1[1]

        cut_value1 = self.cut_planes[0][idim1][cp1]
        cut_value2 = self.cut_planes[1][idim2][cp2]

        print('idim1:', idim1, 'cut_value1:', cut_value1)
        print('idim2:', idim2, 'cut_value2:', cut_value2)

        # Select one possible permutation that will make the second
        # structure being cutted along the same dimension as the first one
        permsel1 = None
        permsel2 = None
        for i in itertools.permutations(range(3), 3):
            if i[idim1] == idim2:
                permsel1 = i
            if i[idim2] == idim1:
                permsel2 = i
            if permsel1 is not None and permsel2 is not None:
                break

        print('Permutation Selected 1:', permsel1)
        print('Permutation Selected 2:', permsel2)

        newreduced1 = np.zeros((self.structures[0].natom, 3))
        newreduced2 = np.zeros((self.structures[1].natom, 3))
        symbols1 = self.structures[0].natom * ['U']
        symbols2 = self.structures[1].natom * ['U']

        nsites = self.structures[0].nsites

        index1 = 0
        index2 = 0

        for i in range(nsites):
            if i in self.split_sites[0][grp0]:
                newreduced1[index1] = self.structures[0].reduced[i, :]
                symbols1[index1] = self.structures[0].symbols[i]
                index1 += 1
            else:
                pos = self.structures[0].reduced[i, permsel1]
                # print i,' - ', self.structures[0].reduced[i], '==>', pos
                newreduced2[index2] = pos
                symbols2[index2] = self.structures[0].symbols[i]
                index2 += 1

        # print '1 Reduced\n', newreduced1
        # print '2 Reduced\n', newreduced2

        # print '1 Symbols:', symbols1
        # print '2 Symbols:', symbols2

        for i in range(nsites):
            if i in self.split_sites[1][grp1]:
                newreduced2[index2] = self.structures[1].reduced[i, :]
                symbols2[index2] = self.structures[1].symbols[i]
                index2 += 1
            else:
                pos = self.structures[1].reduced[i, permsel1]
                # print i,' - ', self.structures[1].reduced[i], '==>', pos
                newreduced1[index1] = pos
                symbols1[index1] = self.structures[1].symbols[i]
                index1 += 1

        cell1 = np.array(self.structures[0].cell)
        cell2 = np.array(self.structures[1].cell)

        # The Splitting is not checking for right stochiometry right now
        # Reverting the symbols to the original ones
        symbols1 = self.structures[0].symbols
        symbols2 = self.structures[1].symbols

        newst1 = Structure(symbols=symbols1, reduced=newreduced1, cell=cell1)
        newst2 = Structure(symbols=symbols2, reduced=newreduced2, cell=cell2)

        return newst1, newst2


def get_cut_planes(structure, delta_distance=0.1):
    """
    Compute all the possible cutting planes defined on reduced coordinates
    for which the structure can be efectively cut considering that the
    perpendicular distance between two atoms is larger than  'delta_distance'

    :param structure:
    :param delta_distance:
    :return:
    """
    ret = {}

    for idim in range(3):
        ret[idim] = []
        # Make a copy of the reduced positions
        v = np.array(structure.reduced[:, idim])
        # Extent the last position with a replica of the first one
        # Because the structure is periodic
        v.sort()
        v = np.concatenate((v, [1 + v[0]]))
        # Compute the vector of differences df[i]=v[i+1]-v[i]
        df = np.diff(v)
        for i in (range(structure.nsites)):
            if df[i] > delta_distance:
                ret[idim].append((v[i] + 0.5 * df[i]) % 1.0)
        ret[idim].sort()
    return ret


def get_simple_split_sites(structure, cut_planes):
    ret = {}
    for idim in range(3):
        for iplane in range(len(cut_planes[idim])):
            ret[(idim, iplane)] = []
            for i in range(structure.natom):
                if structure.reduced[i, idim] < cut_planes[idim][iplane]:
                    ret[(idim, iplane)].append(i)

    # Removing trivial partitions of the entire structure
    for i in ret.keys():
        if len(ret[i]) == structure.nsites:
            ret.pop(i)
    return ret


def get_split_sites(structure, idim, cut_planes, indices):
    """
    Return the indices of the sites corresponding to each
    partition of the structure

    :param structure:
    :param idim:
    :param cut_planes:
    :param indices:
    :return:
    """

    value0 = cut_planes[idim][indices[0]]
    value1 = cut_planes[idim][indices[1]]

    vrange = [[], []]
    if value1 > value0:
        vrange[0] = [(0, value0), (value1, 1)]
        vrange[1] = [(value0, value1)]
    else:
        vrange[0] = [(0, value1), (value0, 1)]
        vrange[1] = [(value1, value0)]

    factor0 = vrange[1][0][1] - vrange[1][0][0]
    factor1 = 1.0 - vrange[0][1][0] + vrange[0][0][1]

    ret = [[], []]
    for i in range(structure.nsites):
        for j in range(2):
            for irange in vrange[j]:
                if irange[0] <= structure.reduced[i, idim] < irange[1]:
                    ret[j].append(i)
    return ret, (factor0, factor1)


def get_all_splitted_compositions(structure, cut_planes):
    ret = []
    for idim in range(3):
        nplanes = len(cut_planes[idim])
        if nplanes > 1:
            for j in itertools.combinations(range(nplanes), 2):
                split_sites, factors = get_split_sites(structure, idim, cut_planes, j)
                npsymbols = np.array(structure.symbols)
                for k in range(2):
                    symbols = npsymbols[split_sites[k]]
                    comp = Composition(symbols)
                    factor = factors[k]
                    ret.append({'dim': idim, 'plane_indices': j, 'composition': comp.composition,
                                'natom': len(symbols), 'factor': factor})
    return ret


def get_matching_options(structure1, structure2, delta_distance=0.1):
    # Get Cutting Planes
    cp1 = get_cut_planes(structure1, delta_distance)
    cp2 = get_cut_planes(structure2, delta_distance)

    # Split Compositions
    sc1 = get_all_splitted_compositions(structure1, cp1)
    sc2 = get_all_splitted_compositions(structure2, cp2)

    # Indexing the Split Compositions
    index_sc1 = {}
    for i in range(len(sc1)):
        natom = sc1[i]['natom']
        if natom in index_sc1:
            index_sc1[natom].append(i)
        else:
            index_sc1[natom] = [i]

    index_sc2 = {}
    for i in range(len(sc2)):
        natom = sc2[i]['natom']
        if natom in index_sc2:
            index_sc2[natom].append(i)
        else:
            index_sc2[natom] = [i]

    ret = []
    for i in range(structure1.nsites / 2 + 1):
        if i in index_sc1 and i in index_sc2:
            for j in index_sc1[i]:
                for k in index_sc2[i]:
                    if sc1[j]['composition'] == sc2[k]['composition']:
                        ret.append({'dim1': sc1[j]['dim'], 'plane_indices1': sc1[j]['plane_indices'],
                                    'dim2': sc2[k]['dim'], 'plane_indices2': sc2[k]['plane_indices']})
    return ret, cp1, cp2


def do_mixing(structure1, structure2, cut_planes1, cut_planes2, matching, match_index):
    # Getting the atoms involved in both partitions:
    dim1 = matching[match_index]['dim1']
    partition1 = get_split_sites(structure1, dim1, cut_planes1, matching[match_index]['plane_indices1'])
    dim2 = matching[match_index]['dim2']
    partition2 = get_split_sites(structure1, dim2, cut_planes2, matching[match_index]['plane_indices2'])

    index1 = matching[match_index]['plane_indices1'][0]
    index2 = matching[match_index]['plane_indices1'][1]

    reduced11 = np.array([structure1.reduced[x] for x in partition1[0]])
    reduced11[:, dim1] = (reduced11[:, dim1] - cut_planes1[dim1][index1] + 1.0) % 1.0
    reduced11[:, dim1] -= np.min(reduced11[:, dim1])
    symbols = [structure1.symbols[x] for x in partition1[0]]
    st11 = Structure(reduced=reduced11, cell=structure1.cell, symbols=symbols)

    reduced12 = np.array([structure1.reduced[x] for x in partition1[1]])
    reduced12[:, dim1] = (reduced12[:, dim1] - cut_planes1[dim1][index2] + 1.0) % 1.0
    reduced12[:, dim1] -= np.min(reduced12[:, dim1])
    symbols = [structure1.symbols[x] for x in partition1[1]]
    st12 = Structure(reduced=reduced12, cell=structure1.cell, symbols=symbols)

    index1 = matching[match_index]['plane_indices2'][0]
    index2 = matching[match_index]['plane_indices2'][1]

    reduced21 = np.array([structure2.reduced[x] for x in partition2[0]])
    reduced21[:, dim2] = (reduced21[:, dim2] - cut_planes2[dim2][index1] + 1.0) % 1.0
    reduced21[:, dim2] -= np.min(reduced21[:, dim2])
    symbols = [structure2.symbols[x] for x in partition2[0]]
    st21 = Structure(reduced=reduced21, cell=structure2.cell, symbols=symbols)

    reduced22 = np.array([structure2.reduced[x] for x in partition2[1]])
    reduced22[:, dim2] = (reduced22[:, dim2] - cut_planes2[dim2][index2] + 1.0) % 1.0
    reduced22[:, dim2] -= np.min(reduced22[:, dim2])
    symbols = [structure2.symbols[x] for x in partition2[1]]
    st22 = Structure(reduced=reduced22, cell=structure2.cell, symbols=symbols)

    return st11, st12, st21, st22



import pymongo


def has_local_mongo():
    try:
        maxSevSelDelay = 1
        client = pymongo.MongoClient("localhost", serverSelectionTimeoutMS=maxSevSelDelay)
        client.server_info()  # force connection on a request as the
        return True
    except pymongo.errors.ServerSelectionTimeoutError as err:
        print(err)
        return False



import os
import numpy as _np
import pychemia

__author__ = "Guillermo Avendano-Franco"
__copyright__ = "Copyright 2016"
__version__ = "0.1"
__maintainer__ = "Guillermo Avendano-Franco"
__email__ = "gtux.gaf@gmail.com"
__status__ = "Development"
__date__ = "May 16, 2016"


def read_kpoints(path='KPOINTS'):
    """
    Load the file KPOINTS in the directory 'path' or
    read directly the file 'path' and return a kpoints
    object for pychemia

    :param path: (str) File path for KPOINTS file
    :return:
    """
    if os.path.isfile(path):
        filename = path
    elif os.path.isdir(path) and os.path.isfile(path + '/KPOINTS'):
        filename = path + '/KPOINTS'
    else:
        print("KPOINTS path not found")
        return

    # Reading the KPOINTS file
    rf = open(filename, 'r')
    comment = rf.readline()
    del comment
    nkpt = int(rf.readline())
    mode = rf.readline()
    if nkpt > 0:
        if mode[0].lower() in ['c', 'k']:
            kmode = 'Cartesian'
        else:
            kmode = 'Reciprocal'
        kp = pychemia.crystal.KPoints(kmode=kmode)
        for i in range(nkpt):
            line = _np.array([float(x) for x in rf.readline().split()])
            pos = line[:3]
            wgt = line[3]
            kp.add_kpt(pos, wgt)

    else:
        if mode[0].lower() in ['g']:
            kmode = 'Gamma'
        elif mode[0].lower() in ['m']:
            kmode = 'Monkhorst-pack'
        else:
            raise ValueError("Kpoints mode must be 'Gamma' or 'Monkhorst-pack'")
        kp = pychemia.crystal.KPoints(kmode=kmode)
        line = _np.array([int(x) for x in rf.readline().split()])
        grid = line[:3]
        try:
            line = _np.array([float(x) for x in rf.readline().split()])
            shift = line[:3]
        except ValueError:
            shift = _np.zeros(3)

        kp.set_grid(grid, shift)

    return kp


def write_kpoints(kp, filepath='KPOINTS'):
    """
    Takes an object kpoints from pychemia and
    save the file KPOINTS in the directory 'path' or
    save the file 'path' as a VASP KPOINTS file

    :param kp: Kpoints object
    :param filepath: (str) Filename where the KPOINTS file is created
    """

    if os.path.isdir(filepath):
        filename = filepath + '/KPOINTS'
    else:
        filename = filepath

    wf = open(filename, 'w')

    wf.write('Automatic mesh\n')
    if kp.kmode == 'cartesian' or kp.kmode == 'reciprocal':
        wf.write(str(kp.nkpt) + '\n')
        wf.write(kp.kmode.title() + '\n')
        for i in range(kp.nkpt):
            wf.write(" %15.7f %15.7f %15.7f %20.7f\n"
                     % (kp.kpoints_list[i, 0],
                        kp.kpoints_list[i, 1],
                        kp.kpoints_list[i, 2],
                        kp.weights[i]))
    elif kp.kmode == 'gamma' or kp.kmode == 'monkhorst-pack':
        wf.write('0\n')
        wf.write(kp.kmode.title() + '\n')
        wf.write(" %7d %7d %7d\n" % tuple(kp.grid))
        wf.write(" %7.4f %7.4f %7.4f\n" % tuple(kp.shifts))

    wf.close()



from pychemia import HAS_PYMONGO
from .realfunction import RealFunction

if HAS_PYMONGO:
    from .relaxstructures import RelaxStructures
    from .orbitaldftu import OrbitalDFTU
    from .noncollinearmagmoms import NonCollinearMagMoms
    from .ljcluster import LJCluster

from . import orbitaldftu

"""
Populations are objects collecting structures/properties usually for feeding global search algorithms.
Several populations are implemented.

Several population classes are defined

__RelaxStructures__: Used for structural search.
__OrbitalDFTU__: Used for optimization of orbital mixing and determination of U on DFT+U calculations.
__NonCollinearMagMoms__: Optimize the directions of MagMoms for Non-Collinear-Spin calculations.
__RealFunction__: A population of vector for optimize real-valuated functions.
__LJCluster__: Stores Lennard-Jones clusters.

"""


from __future__ import unicode_literals
import json
import numpy as np
from builtins import str
from abc import ABCMeta, abstractmethod
from pychemia import HAS_PYMONGO
from pychemia.utils.computing import deep_unicode

if HAS_PYMONGO:
    from pychemia.db import PyChemiaDB


class Population:
    __metaclass__ = ABCMeta
    """
    General class for all optimization algorithms that uses fixed and blocked
    Generations
    """

    def __init__(self, name, tag, use_mongo=True, direct_evaluation=False):

        name = deep_unicode(name)
        self.tag = tag
        self.pcdb = None
        self.direct_evaluation = direct_evaluation
        if isinstance(name, str):
            self.name = name
            if use_mongo:
                self.pcdb = PyChemiaDB(name)
        else:
            self.name = name.name
            if use_mongo:
                self.pcdb = name

    def __iter__(self):
        return self.pcdb.entries.find()

    def __len__(self):
        return len(self.members)

    def __str__(self):
        ret = ' Population Name:     %s\n' % self.name
        ret += ' Tag:                 %s\n' % self.tag
        ret += ' Members:             %s\n' % len(self)
        return ret

    def disable(self, entry_id):
        self.pcdb.entries.update({'_id': entry_id}, {'$set': {'status.' + self.tag: False}})

    def enable(self, entry_id):
        self.pcdb.entries.update({'_id': entry_id}, {'$set': {'status.' + self.tag: True}})
        if self.direct_evaluation:
            self.evaluate_entry(entry_id)

    def get_values(self, selection):
        ret = {}
        for i in selection:
            ret[i] = self.value(i)
        return ret

    def update_properties(self, entry_id, new_properties):
        self.pcdb.update(entry_id, properties=new_properties)

    def set_in_properties(self, entry_id, field, value):
        return self.pcdb.entries.update_one({'_id': entry_id}, {'$set': {'properties.'+field: value}})

    def get_population_info(self):
        return self.pcdb.db.population_info.find_one({'tag': self.tag})

    def insert_entry(self, entry):
        if 'structure' not in entry:
            entry['structure'] = {}
        if 'properties' not in entry:
            entry['properties'] = {}
        if 'status' not in entry:
            entry['status'] = {}
        return self.pcdb.entries.insert(entry)

    def set_entry(self, entry):
        return self.insert_entry(entry)

    def get_structure(self, entry_id):
        return self.pcdb.get_structure(entry_id)

    def set_structure(self, entry_id, structure):
        return self.pcdb.update(entry_id, structure=structure)

    def unset_properties(self, entry_id):
        return self.pcdb.update(entry_id, properties={})

    def get_entry(self, entry_id, projection=None, with_id=True):
        """
        Return an entry identified by 'entry_id'

        :param with_id:
        :param projection: Insert that projection into the query
        :param entry_id: A database identifier
        :return:
        """
        if not with_id:
            projection['_id'] = 0
        entry = self.pcdb.entries.find_one({'_id': entry_id}, projection)
        return entry

    def ids_sorted(self, selection):
        values = np.array([self.value(i) for i in selection])
        sorted_indices = np.argsort(values)
        return np.array(selection)[sorted_indices]

    def load_json(self, filename):
        filep = open(filename, 'r')
        data = json.load(filep)
        for entry in data:
            self.pcdb.entries.insert(entry)

    def random_population(self, n):
        """
        Create N new random structures to the population

        :param n: (int) The number of new structures
        :return: (list) The identifiers for the new structures
        """
        return [self.add_random() for i in range(n)]

    def replace_failed(self):
        pass

    def save_info(self):
        data = self.pcdb.db.population_info.find_one({'_id': self.tag})
        if data is None:
            data = self.to_dict
            data['_id'] = self.tag
            self.pcdb.db.population_info.insert(data)
        else:
            self.pcdb.db.population_info.update({'_id': self.tag}, self.to_dict)

    def save_json(self, filename):
        ret = []
        for entry_id in self.members:
            ret.append(self.get_entry(entry_id, with_id=False))
        filep = open(filename, 'w')
        json.dump(ret, filep, sort_keys=True, indent=4, separators=(',', ': '))

    def unlock_all(self, name=None):
        for i in self.members:
            self.pcdb.unlock(i, name=name)

    @abstractmethod
    def add_random(self):
        pass

    @abstractmethod
    def check_duplicates(self, ids):
        pass

    @abstractmethod
    def cross(self, ids):
        pass

    @abstractmethod
    def distance(self, entry_id, entry_jd):
        pass

    @abstractmethod
    def evaluate_entry(self, entry_id):
        pass

    @abstractmethod
    def get_duplicates(self, ids):
        pass

    @abstractmethod
    def from_dict(self, population_dict):
        pass

    @abstractmethod
    def is_evaluated(self, entry_id):
        pass

    @abstractmethod
    def move_random(self, entry_id, factor=0.2, in_place=False, kind='move'):
        pass

    @abstractmethod
    def move(self, entry_id, entry_jd, factor=0.2, in_place=False):
        pass

    @abstractmethod
    def new_entry(self, data, active=True):
        pass

    @abstractmethod
    def recover(self):
        pass

    @abstractmethod
    def value(self, entry_id):
        pass

    @abstractmethod
    def str_entry(self, entry_id):
        pass

    @property
    def actives(self):
        return [entry['_id'] for entry in self.pcdb.entries.find({'status.' + self.tag: True}, {'_id': 1})]

    @property
    def actives_evaluated(self):
        return [x for x in self.actives if self.is_evaluated(x)]

    @property
    def actives_no_evaluated(self):
        return [x for x in self.actives if not self.is_evaluated(x)]

    @property
    def evaluated(self):
        return [entry for entry in self.members if self.is_evaluated(entry)]

    @property
    def fraction_evaluated(self):
        ret = np.sum([1 for i in self.actives if self.is_evaluated(i)])
        return float(ret) / len(self.actives)

    @property
    def members(self):
        return [x['_id'] for x in self.pcdb.entries.find({}, {'_id': 1})]

    @property
    def to_dict(self):
        return {'name': self.name, 'tag': self.tag}

    @property
    def best_candidate(self):
        if len(self.evaluated) > 0:
            return self.ids_sorted(self.evaluated)[0]
        else:
            return None

    def refine_progressive(self, entry_id):
        pass



import unittest
import numpy as np
from pychemia.utils.serializer import generic_serializer


class SerializerTest(unittest.TestCase):

    def test_serializer(self):
        """
        Tests (pychemia.utils.serializer)                            :
        """
        a = np.array([1, 2, 3])
        assert generic_serializer(a) == [1, 2, 3]
        b = np.array([[1, 2, 3], [4, 5, 6]])
        assert generic_serializer(b) == [[1, 2, 3], [4, 5, 6]]
        c = np.array([[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [0, 1, 2]]])
        assert generic_serializer(c) == [[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [0, 1, 2]]]
        mydict = {'a': a, 'b': b, 'c': c}
        generic_serializer(mydict) == {u'a': [1, 2, 3],
                                       u'b': [[1, 2, 3], [4, 5, 6]],
                                       u'c': [[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [0, 1, 2]]]}



"""
Set of classes and functions to manipulate

OCTOPUS 'inp' input files

"""
from pychemia import HAS_SCIPY
import numpy as np
from .input import *

if HAS_SCIPY:
    from .analysis import *


# __all__ = filter(lambda s: not s.startswith('_'), dir())


def coordinates(dirname):
    """
    Reads the 'coordinates' file and extracts its
    data as the numpy arrays 'positions','velocities'
    and 'forces'.

    Each of those arrays has the following structure:

    Args:
       dirname:
          The path were the octopus data is produced

    Returns:
       natom:
          Number of atoms
       iterations:
          Array with iterations
       times:
          Array with times
       positions:
          positions[0:MaxIteration,0:natom,0:3]
       velocities:
          velocities[0:MaxIteration,0:natom,0:3]
       forces:
          forces[0:MaxIteration,0:natom,0:3]

    """

    geo = np.loadtxt(dirname + '/td.general/coordinates')
    natom = (geo.shape[1] - 2) / 9
    iterations = geo[:, 0]
    times = geo[:, 1]

    positions = geo[:, 2:3 * natom + 2]
    velocities = geo[:, 3 * natom + 2:6 * natom + 2]
    forces = geo[:, 6 * natom + 2:9 * natom + 2]

    positions = positions.reshape([-1, 3])
    velocities = velocities.reshape([-1, 3])
    forces = forces.reshape([-1, 3])

    return natom, iterations, times, positions, velocities, forces


def energy(dirname):
    """
    Reads the 'energy' file and extracts its
    data as the numpy arrays 'Total','Kinetic', etc.

    Args:
       dirname:
          The directory where the octopus data was produced

    Returns:
       iterations:
          numpy array of iterations recorded
       times:
          numpy array with time for each iteration
       energy_dict:
          dictionary of numpy arrays with values of
          Total, Kinetic (ions), Ion-Ion, Electronic,
          Eigenvalues, Hartree, Int[n v_xc], Exchange
          and Correlation

    """

    ene = np.loadtxt(dirname + '/td.general/energy')
    iterations = ene[:, 0]
    times = ene[:, 1]

    energy_dict = {'Total': ene[:, 2], 'Kinetic': ene[:, 3], 'Ion-Ion': ene[:, 4], 'Electronic': ene[:, 5],
                   'Eigenvalues': ene[:, 6], 'Hartree': ene[:, 7], 'Int': ene[:, 8], 'Exchange': ene[:, 9],
                   'Correlation': ene[:, 10]}

    return iterations, times, energy_dict



import unittest
import time
import multiprocessing
from pychemia.db import has_connection
from pychemia import pcm_log
from pychemia.searcher import HarmonySearch, FireFly, GeneticAlgorithm, ParticleSwarm
from pychemia.population import LJCluster


def evaluator():
    from pychemia.evaluator import cluster_launcher
    dbsettings = {'name': 'test'}
    cluster_launcher(dbsettings, 4)


def searcher():
    popu = LJCluster('test', 'Ar13', refine=False, minimal_density=40)
    popu.pcdb.clean()
    hs = HarmonySearch(popu, generation_size=8, stabilization_limit=3)
    hs.run()


def notest_searcher():
    """
    Testing HarmonySearch               :
    """
    if not has_connection():
        return

    p1 = multiprocessing.Process(target=searcher)
    p2 = multiprocessing.Process(target=evaluator)

    p1.start()
    time.sleep(10)
    p2.start()


class SearcherTest(unittest.TestCase):

    def test_firefly(self):
        """
        Tests (pychemia.searcher.firefly) with LJ Clusters           :
        """
        if not has_connection():
            return
        pcm_log.debug('FireFly')
        popu = LJCluster('test', composition='Xe13', refine=True, direct_evaluation=True)
        popu.pcdb.clean()
        searcher = FireFly(popu, generation_size=8, stabilization_limit=3)
        searcher.run()
        popu.pcdb.clean()
        searcher = FireFly(popu, {'delta': 0.1, 'gamma': 0.1, 'beta0': 0.8, 'alpha0': 0, 'multi_move': True},
                           generation_size=8, stabilization_limit=3)
        searcher.run()
        popu.pcdb.clean()

    def test_genetic(self):
        """

        Tests (pychemia.searcher.genetic) with LJ Clusters           :
        """
        if not has_connection():
            return
        pcm_log.debug('GeneticAlgorithm')
        popu = LJCluster('test', composition='Xe13', refine=False, direct_evaluation=True)
        popu.pcdb.clean()
        searcher = GeneticAlgorithm(popu, generation_size=8, stabilization_limit=3)
        searcher.run()
        popu.pcdb.clean()

    def test_harmony(self):
        """
        Tests (pychemia.searcher.harmony) with LJ Clusters           :
        """
        if not has_connection():
            return
        pcm_log.debug('HarmonySearch')
        popu = LJCluster('test', composition='Xe13', refine=False, direct_evaluation=True)
        popu.pcdb.clean()
        searcher = HarmonySearch(popu, generation_size=8, stabilization_limit=3)
        searcher.run()
        popu.pcdb.clean()

    def test_swarm(self):
        """
        Tests (pychemia.searcher.swarm) with LJ Clusters             :
        """
        if not has_connection():
            return
        pcm_log.debug('ParticleSwarm')
        popu = LJCluster('test', composition='Xe13', refine=False, direct_evaluation=True)
        popu.pcdb.clean()
        searcher = ParticleSwarm(popu, generation_size=8, stabilization_limit=3)
        searcher.run()
        popu.pcdb.clean()




import os
import subprocess
from multiprocessing import Process
import time


class Runner:
    def __init__(self, code, code_bin, environment, use_mpi=True, nmpiproc=2, nconcurrent=1, runtime=3600):

        if code.lower() not in ['abinit', 'vasp', 'octopus', 'dftb', 'fireball']:
            raise ValueError('Code not supported: ', code)
        else:
            self.code = code.lower()

        if environment.lower() not in ['local', 'remote']:
            raise ValueError('Environment must be local or remote: ')
        else:
            self.environment = environment.lower()

        self.use_mpi = use_mpi
        self.nmpiproc = nmpiproc
        self.nconcurrent = nconcurrent
        self.code_bin = code_bin
        self.runtime = runtime

    def initialize(self, dirpath):
        """
        Utility that copy a given script and execute the given
        command inside the directory

        :param dirpath: (str) Directory to execute runner
        """
        if not os.path.isdir(dirpath):
            os.mkdir(dirpath)
        if self.code == 'abinit':
            pass
            # pychemia.code.abinit.AbiFiles(basedir=dirpath)
        elif self.code == 'vasp':
            pass

    def run(self, dirpath='.', analyser=None):

        if self.code == 'abinit':
            outfile = 'abinit.stdout'
            errfile = 'abinit.stderr'
            infile = 'abinit.in'
        elif self.code == 'vasp':
            outfile = 'vasp.stdout'
            errfile = 'vasp.stderr'
            infile = None

        if self.environment == 'local':

            def worker(path):
                cwd = os.getcwd()
                os.chdir(path)
                outf = open(outfile, 'a')
                errf = open(errfile, 'a')
                for i in [outf, errf]:
                    i.write('' + (40 * '=') + ' New Run ' + (40 * '=') + '\n')
                if infile is not None:
                    rf = open(infile, 'r')
                else:
                    rf = None

                initime = time.time()
                if self.use_mpi:
                    childp = subprocess.Popen(['mpirun', '-np', str(self.nmpiproc),
                                               '--map-by', 'socket:PE=2', self.code_bin],
                                              stdin=rf, stdout=outf, stderr=errf)
                else:
                    childp = subprocess.Popen([self.code_bin], stdin=rf, stdout=outf, stderr=errf)

                childp.poll()
                while childp.returncode is None:
                    childp.poll()
                    rf = open('vasp.stdout', 'r')
                    for iline in rf.readlines()[-20:]:
                        if iline.strip().startswith("WARNING: Sub-Space-Matrix is not hermitian in DAV"):
                            print("[WARNING: Sub-Space-Matrix is not hermitian in DAV] Stopping now...")
                            self._stop_run(childp)
                    if time.time() - initime > self.runtime:
                        self._stop_run(childp)
                        break
                    else:
                        if analyser is not None:
                            ret = analyser()
                            print(os.path.basename(path), ret)
                        time.sleep(60)

                print('The return code was', childp.returncode)
                os.chdir(cwd)
                errf.close()
                outf.close()

            p = Process(target=worker, args=(dirpath,))
            p.start()
            p.join()

        return p

    @staticmethod
    def _stop_run(childp, softtime=600, extratime=60):
        print('Creating Stoping files')
        wf = open('CHKPT', 'w')
        wf.close()
        wf = open('STOPCAR', 'w')
        wf.write('LSTOP = .TRUE.\n')
        wf.close()
        time.sleep(softtime)

        childp.poll()
        if childp.returncode is None:
            print('Sending SIGTERM to process')
            childp.terminate()
            time.sleep(extratime)

        childp.poll()
        if childp.returncode is None:
            print('Sending SIGKILL to process')
            childp.kill()
        if os.path.isfile('CHKPT'):
            os.remove('CHKPT')
        if os.path.isfile('STOPCAR'):
            os.remove('STOPCAR')

    def run_multidirs(self, workdirs, worker, checker):
        pt = []
        for i in range(self.nconcurrent):
            pt.append(None)

        index = 0
        while True:
            for i in range(self.nconcurrent):
                if pt[i] is None or not pt[i].is_alive():
                    ret = checker(workdirs[index])
                    if ret:
                        print('Launching for process ' + str(i) + ' on dir ' + os.path.basename(
                            workdirs[index]) + ' index ' + str(index))
                        pt[i] = Process(target=worker, args=(workdirs[index],))
                        pt[i].start()
                    index = (index + 1) % len(workdirs)
            time.sleep(30)
            complete = True
            for idir in workdirs:
                if not os.path.isfile(idir + os.sep + 'COMPLETE'):
                    complete = False
            if complete:
                print('Finishing...')
                break

    def run_multidirs_nonstop(self, workdirs, worker, checker):

        def superworker(s_workdirs, s_worker, s_checker):
            self.run_multidirs(s_workdirs, s_worker, s_checker)

        proc = Process(target=superworker, args=(workdirs, worker, checker))
        proc.start()



import numpy as np
from pychemia import pcm_log
import scipy.optimize
import scipy.spatial
from pychemia.utils.mathematics import length_vectors
from pychemia.utils.periodic import mass


class LennardJones:
    def __init__(self, structure, ljparams=None, cp=0.0):
        self.initial_structure = structure
        self.structure = structure.copy()

        if ljparams is None:
            ljparams = {}
            for i in self.structure.species:
                for j in self.structure.species:
                    ljparams[i + '-' + j] = {'sigma': 1.0, 'epsilon': 1.0}
        self.ljparams = ljparams
        self.cp = cp
        self.sigmas = np.zeros((structure.natom, structure.natom))
        self.epsilons = np.zeros((structure.natom, structure.natom))
        for i in range(structure.natom):
            for j in range(structure.natom):
                self.sigmas[i, j], self.epsilons[i, j] = self.get_sigma_epsilon(i, j)

    def get_forces(self):
        return lj_forces(self.structure.positions, self.sigmas, self.epsilons, self.cp).reshape((-1, 3))

    def get_magnitude_forces(self):
        return length_vectors(self.get_forces())

    def steepest_descent(self, dt=1, tolerance=1E-3):
        forces = self.get_forces()
        while np.max(self.get_magnitude_forces()) > tolerance:
            for i in range(self.structure.natom):
                imass = mass(self.structure.symbols[i])
                self.structure.positions[i] += 0.5 * forces[i] / imass * dt ** 2

            forces = self.get_forces()
            maxforce = np.max(self.get_magnitude_forces())
            dt = max(100.0 / maxforce, 1.0)
            pcm_log.debug('Steepest Descent maxforce= %9.3E   density=%7.4f dt=%7.3f' % (maxforce,
                                                                                         self.structure.density, dt))

    def get_sigma_epsilon(self, i, j):
        specie1 = self.structure.symbols[i]
        specie2 = self.structure.symbols[j]
        sort_species = [specie1, specie2]
        sort_species.sort()

        sigma = self.ljparams[sort_species[0] + '-' + sort_species[1]]['sigma']
        epsilon = self.ljparams[sort_species[0] + '-' + sort_species[1]]['epsilon']

        return float(sigma), float(epsilon)

    def get_energy(self):
        return lj_energy(self.structure.positions, self.sigmas, self.epsilons, self.cp)

    def local_minimization(self, method='BFGS', gtol=1E-4, soft_max_ncalls=5, hard_max_ncalls=20):

        if method in ['Nelder-Mead', 'Powell']:
            jac = None
            options = {'maxiter': 100, 'disp': False}
        else:
            jac = lj_gradient
            options = {'gtol': 0.1 * gtol, 'disp': False}

        x0 = self.structure.positions.flatten()

        hard_call = 1
        soft_call = 1
        res = 0.0

        while True:

            res = scipy.optimize.minimize(lj_energy, x0, args=(self.sigmas, self.epsilons, self.cp), method=method,
                                          jac=jac, options=options)

            if method in ['BFGS', 'CG']:
                maxforce = np.max(np.apply_along_axis(np.linalg.norm, 1, np.array(res.jac).reshape((-1, 3))))
                pcm_log.debug('[%2d/%2d] MaxForce/gtol= %9.3E/%9.3E  Value= %12.5f' % (soft_call,
                                                                                       soft_max_ncalls,
                                                                                       maxforce, gtol,
                                                                                       res.fun))

                x0 = res.x
                soft_call += 1
                hard_call += 1
                if maxforce < gtol:
                    break
                if soft_call > soft_max_ncalls:
                    pcm_log.debug('Anomalous condition, maxforce= %9.3E gtol= %9.3E' % (maxforce, gtol))
                    pos = np.array(x0).reshape((-1, 3))
                    mindis = np.min(np.array(np.array(scipy.spatial.distance_matrix(pos, pos)) +
                                             100 * np.eye(len(pos))).flatten())
                    x0 = np.array(1.0 / mindis * pos).flatten()
                    soft_call = 1
                if hard_call > hard_max_ncalls:
                    pcm_log.debug('Could not reach target forces, maxforce= %9.3E gtol= %9.3E' % (maxforce, gtol))
                    pcm_log.debug('Hard limit reached, aborting local minimization')
                    break
            else:
                break
        return res

    def move_to_minima(self):
        relax = self.local_minimization()
        self.structure.positions = relax.x.reshape((-1, 3))
        self.structure.relocate_to_cm()


def lj_forces(pos, sigmas, epsilons, cp=0.0):
    pos = np.array(pos).reshape((-1, 3))
    n = len(pos)
    ret = np.zeros((n, 3))
    for i in range(n - 1):
        for j in range(i + 1, n):
            vector = pos[j] - pos[i]
            distance = np.linalg.norm(vector)
            if distance < 1E-8:
                print('ERROR: Too small distance between atoms')
                distance = 1E-8
            sr6 = (sigmas[i, j] / distance) ** 6
            magnitude = 24 * epsilons[i, j] / distance * (2.0 * sr6 ** 2 - sr6)
            ret[i] -= magnitude * vector / distance
            ret[j] += magnitude * vector / distance
    if cp > 0.0:
        for i in range(n):
            ret[i] -= cp * pos[i]
    return ret.flatten()


def lj_gradient(pos, sigmas, epsilons, cp=0.0):
    return -lj_forces(pos, sigmas, epsilons, cp)


def lj_energy(pos, sigmas, epsilons, cp=0.0):
    pos = np.array(pos).reshape((-1, 3))
    ret = 0
    n = len(pos)
    for i in range(n - 1):
        for j in range(i + 1, n):
            vector = pos[j] - pos[i]
            distance = np.linalg.norm(vector)
            if distance > 1E-10:
                sr6 = (sigmas[i, j] / distance) ** 6
            else:
                sr6 = 0
            ret += 4 * epsilons[i, j] * (sr6 ** 2 - sr6)
    if cp > 0.0:
        for i in range(n):
            ret += 0.5 * cp * np.linalg.norm(pos[i]) ** 2
    return ret


def lj_compact_evaluate(structure, gtol, minimal_density):
    k = 1
    while structure.density < minimal_density:
        iniden = structure.density
        lj = LennardJones(structure, cp=k)
        relax = lj.local_minimization(gtol=gtol)
        structure.set_positions(relax.x.reshape((-1, 3)))
        finden = structure.density
        pcm_log.debug('Compacting Cluster (target_density=%7.3f): Density I= %7.3f   F= %7.3f' % (minimal_density,
                                                                                                  iniden, finden))
        k += 1
        if k > 10:
            pcm_log.debug('I tried too much...')
            break

    lj = LennardJones(structure)
    relax = lj.local_minimization(gtol=gtol)
    # Return relaxed positions, forces and energy
    return relax.x.reshape((-1, 3)), relax.jac.reshape((-1, 3)), relax.fun



from __future__ import unicode_literals
from numpy import array, argsort
from fractions import gcd as _gcd
from math import pi
from pychemia.utils.periodic import atomic_symbols, electronegativity, atomic_number, covalent_radius
from pychemia.utils.computing import deep_unicode
from builtins import str
from functools import reduce
from collections import Mapping


class Composition(Mapping):
    u"""
    The class Composition is basically a dictionary with species as keys and
    number of atoms of that specie as values. The methods provided for Composition objects should
    not contain geometrical information or graph connectivity.

    The main purpose of this class is to be able to parse formulas into compositions and return
    string formulas sorted in various ways.
    """

    def __init__(self, value=None):
        """
        Creates a new composition, internally it is a dictionary
        where each specie is the key and the value is an integer
        with the number of atoms of that specie

        :param value: (str, dict) The value could be a string with a chemical formula or the actual dictionary
        of species and values

        :rtype: Composition

        Example:
        >>> import pychemia
        >>> comp = pychemia.Composition({'Ba': 2, 'Cu': 3, 'O': 7, 'Y': 1})
        >>> comp.formula
        u'Ba2Cu3O7Y'
        >>> comp = pychemia.Composition('Ba2Cu3O7Y')
        >>> comp2 = pychemia.Composition(comp)
        >>> len(comp2)
        4
        >>> comp.nspecies
        4
        >>> comp = pychemia.Composition()
        >>> comp.composition
        {}
        >>> len(comp)
        0

        """
        self._composition = {}
        if value is not None:
            value = deep_unicode(value)
        if isinstance(value, str):
            self._set_composition(self.formula_parser(value))
        elif isinstance(value, dict):
            self._set_composition(value)
        elif isinstance(value, Composition):
            self._set_composition(value.composition)
        elif hasattr(value, "__len__"):
            dvalue = {}
            for i in value:
                if i in dvalue:
                    dvalue[i] += 1
                else:
                    dvalue[i] = 1
            self._set_composition(dvalue)
        else:
            self._composition = {}

    def __len__(self):
        return len(self._composition)

    def __getitem__(self, item):
        if item in self._composition:
            return self._composition[item]
        else:
            return 0

    def __repr__(self):
        return 'Composition(' + str(self.composition) + ')'

    def __str__(self):
        ret = ''
        for i in self.species:
            ret += " %3s: %4d  " % (i, self.composition[i])
        return ret

    def __iter__(self):
        return iter(self.composition)

    def __contains__(self, item):
        return item in self._composition

    def _set_composition(self, value):
        """
        Checks the values of a dictionary before seting the actual composition

        :param value: (dict)
        :rtype: None
        """
        for i in value:
            assert (i in atomic_symbols)
            assert (isinstance(value[i], int))
        self._composition = value.copy()

    @property
    def composition(self):
        """
        :return: The composition dictionary

        :rtype: dict
        """
        return self._composition

    @property
    def formula(self):
        """
        :return: The chemical formula with atoms sorted alphabetically

        :rtype: str
        """
        return self.sorted_formula(sortby='alpha', reduced=True)

    @property
    def gcd(self):
        """
        The number of formulas that can be extracted from a composition
        The greatest common denominator for the composition.

        :rtype: (int)

        Example:
        >>> import pychemia
        >>> comp = pychemia.Composition('NaCl')
        >>> comp.gcd
        1
        >>> comp = pychemia.Composition('Na2Cl2')
        >>> comp.gcd
        2
        >>> comp = pychemia.Composition()
        >>> comp.gcd is None
        True

        """
        if self.natom > 0:
            return reduce(_gcd, self.values)
        else:
            return None

    @property
    def symbols(self):
        ret = []
        for specie in self:
            number_atoms_specie = self.composition[specie]
            for i in range(number_atoms_specie):
                ret.append(specie)
        return sorted(deep_unicode(ret))

    @property
    def species(self):
        """
        :return: The list of species

        :rtype: list
        """
        return [deep_unicode(x) for x in self._composition]

    @property
    def nspecies(self):
        return len(self.species)

    @property
    def values(self):
        """
        :return: The number of atoms of each specie

        :rtype: list
        """
        return [self._composition[x] for x in self._composition]

    @property
    def natom(self):
        """
        :return: The number of atoms in the composition

        :rtype: int
        """
        return sum(self.values)

    @staticmethod
    def formula_parser(value):
        """
        :return: Convert an string representing a chemical formula into a dictionary with the species as keys
                 and values as the number of atoms of that specie

        :param value: (str) String representing a chemical formula

        :rtype: dict

        Examples:
        >>> import pychemia
        >>> import pprint
        >>> pychemia.Composition.formula_parser('Au20')
        {u'Au': 20}
        >>> ret = pychemia.Composition.formula_parser('UutUupUusUuo')
        >>> pprint.pprint(ret)
        {u'Uuo': 1, u'Uup': 1, u'Uus': 1, u'Uut': 1}

        """
        ret = {}
        jump = False
        for i in range(len(value)):
            if jump > 0:  # This char belongs to the current atom, move on
                jump -= 1
            elif value[i].isupper():  # Atom Name starts with Uppercase
                if i + 1 < len(value) and value[i + 1].islower():  # Atom name has more than 1 char
                    if i + 2 < len(value) and value[i + 2].islower():  # Atom name has more than 2 chars
                        specie = value[i:i + 3]
                        jump = 2
                    else:
                        specie = value[i:i + 2]
                        jump = 1
                else:
                    specie = value[i]
                    jump = 0
                j = 1
                number = ''
                while True:
                    if i + jump + j < len(value) and value[i + jump + j].isdigit():
                        number += value[i + jump + j]
                        j += 1
                    else:
                        break
                if number == '':
                    ret[specie] = 1
                else:
                    ret[specie] = int(number)
        return ret

    @staticmethod
    def formula_to_list(formula, nunits=1):
        """
        Reads a formula and returns a list of
        atomic symbols consistent with the formula
        and the number of formulas given by nunits

        :param formula: (str) Chemical formula as string

        :param nunits: (int) Number of formulas to apply

        :rtype : (list)

        Examples:
        >>> import pychemia
        >>> pychemia.Composition.formula_to_list('NaCl')
        [u'Na', u'Cl']
        >>> flist = pychemia.Composition.formula_to_list(u'Uut2Uup3Uus4Uuo5')
        >>> len(flist)
        14
        >>> flist = pychemia.Composition.formula_to_list('Uut2Uup3Uus4Uuo5', nunits=2)
        >>> len(flist)
        28

        """
        import re

        # decompose composition
        a = re.findall(r"[A-Z][a-z0-9]*", formula)
        composition = []
        for i in a:
            m = re.match(r"([A-Za-z]+)([0-9]*)", i)
            if m.group(2) == "":
                n = int(1)
            else:
                n = int(m.group(2))

            for j in range(n * nunits):
                composition.append(m.group(1))

        return composition

    def sorted_formula(self, sortby='alpha', reduced=True):
        """
        :return: The chemical formula. It could be sorted  alphabetically using sortby='alpha', by electronegativity
                 using sortby='electroneg' or using Hill System with sortby='Hill'

        :param sortby: (str) 'alpha' : Alphabetically
                             'electroneg' : Electronegativity
                             'hill' : Hill System

        :param reduced: (bool) If the formula should be normalized

        :rtype: str

        >>> comp=Composition('YBa2Cu3O7')
        >>> comp.sorted_formula()
        u'Ba2Cu3O7Y'
        >>> comp.sorted_formula(sortby='hill')
        u'Ba2Cu3O7Y'
        >>> comp.sorted_formula(sortby='electroneg')
        u'Ba2YCu3O7'
        >>> comp = Composition('H10C5')
        >>> comp.sorted_formula(sortby='hill', reduced=True)
        u'CH2'
        >>> comp = Composition('IBr')
        >>> comp.sorted_formula(sortby='hill', reduced=False)
        u'BrI'
        >>> comp = Composition('Cl4C')
        >>> comp.sorted_formula(sortby='hill', reduced=False)
        u'CCl4'
        >>> comp = Composition('IH3C')
        >>> comp.sorted_formula(sortby='hill', reduced=False)
        u'CH3I'
        >>> comp = Composition('BrH5C2')
        >>> comp.sorted_formula(sortby='hill', reduced=False)
        u'C2H5Br'
        >>> comp = Composition('S04H2')
        >>> comp.sorted_formula(sortby='hill', reduced=False)
        u'H2S4'
        >>> comp = Composition('SO4H2')
        >>> comp.sorted_formula(sortby='hill', reduced=False)
        u'H2O4S'

        """
        if reduced and self.gcd > 1:
            comp = Composition(self.composition)
            for i in comp.composition:
                comp._composition[i] //= self.gcd
        else:
            comp = self
        if sortby == 'electroneg':
            electroneg = list(electronegativity(comp.species))
            for i in range(len(electroneg)):
                if electroneg[i] is None:
                    electroneg[i] = -1
            sortedspecies = array(comp.species)[argsort(electroneg)]
        elif sortby == "hill":  # FIXME: Hill system exceptions not implemented
            sortedspecies = []
            presortedspecies = sorted(comp.species)
            if 'C' in presortedspecies:
                sortedspecies.append('C')
                presortedspecies.pop(presortedspecies.index('C'))
            if 'H' in presortedspecies:
                sortedspecies.append('H')
                presortedspecies.pop(presortedspecies.index('H'))
            sortedspecies += presortedspecies
        else:
            sortedspecies = sorted(comp.species)
        ret = u''
        for specie in sortedspecies:
            ret += '%s' % specie
            if comp.composition[specie] > 1:
                ret += "%d" % comp.composition[specie]
        return deep_unicode(ret)

    def species_encoded(self, base):
        ret = 0
        i = 0
        for atom_number in sorted(atomic_number(self.species)):
            ret += atom_number * (base ** i)
            i += 1
        return ret

    def species_hex(self):
        """
        Encodes the species into a hexadecimal representation where
        each specie is stored on a 2-Byte slot ordered by atomic
        number.
        This is a 'confortable' encoding where each 2 characters
        from the hexadecimal will encode a single species and the
        species are ordered by atomic number making the codification
        unique.

        :return: str

        Example:
        >>> comp = Composition('YBa2Cu3O7')
        >>> comp.species_hex()
        '0x38271d08'

        """
        enc = self.species_encoded(256)
        return hex(enc)

    @staticmethod
    def get_species_from_hex(arg):
        """
        Return a set of species from the encoded species hexadecimal
        representation.

        :param arg: str String with hexadecimal representation of list of species.
        :return:

        Example:
        >>> Composition.get_species_from_hex('0x38271d08')
        [8, 29, 39, 56]

        """
        num = int(arg, 16)
        ret = []
        while num > 0:
            ret.append(num % 256)
            num = (num-ret[-1])//256
        return ret

    def covalent_volume(self, packing='cubes'):
        """
        Returns the volume occupied by a given formula
        assuming a 'cubes' packing or 'spheres' packing

        :param packing: (str) The kind of packing could be 'cubes' or 'spheres'

        :rtype : (float)

        >>> import pychemia
        >>> comp=pychemia.Composition('C5H10')
        >>> comp.covalent_volume()
        19.942320000000002
        >>> comp.covalent_volume(packing='spheres')
        10.441774334589468

        """
        if packing == 'cubes':
            factor = 8
        elif packing == 'spheres':
            factor = 4 * pi / 3.0
        else:
            raise ValueError('Non-valid packing value ', packing)

        # find volume of unit cell by adding cubes
        volume = 0.0
        for specie in self:
            number_atoms_specie = self.composition[specie]
            # Pack each atom in a cube (2*r)^3
            volume += factor * number_atoms_specie * covalent_radius(specie) ** 3
        return volume
